Require Relations_2.
Require RAMs.
Require Lt.

(* Example 1: program that add the registers 0 and 1, storing the resul in *)
(* register 0.                                                             *)

Definition example1_program :=
 (cons (DecJump (1) (3))
  (cons (Succ (0))
   (cons (DecJump (2) (0))
    (nil ?)))).

Lemma example1_program_ok: (AllS (uses_at_most (3)) example1_program).
Proof.
 Unfold example1_program;
 Do 3 (* 3 registers *)
  (Constructor;
   [ Constructor;
     Auto
   | Idtac
   ]);
 Constructor.
Qed.

Definition example1_RAM :=
 (Build_RAM (3) example1_program example1_program_ok).

(* 2 + 3 *)
Definition example1_initial_state :=
 [registers :=
   [r:register]
    Cases r of
       O => (S (S O))
     | (S O) => (S (S (S O)))
     | _ => O
    end
 ] (Build_configuration O registers).

(* We define the final state up to extensional equality on *)
(* the registers field                                      *)
Definition example1_final_state :=
 [registers :=
   [r:register]
    Cases r of
       O => (S (S (S (S (S O)))))
     | (S O) => O
     | _ => O
    end
 ] (Build_configuration (S (S (S O))) registers).

(* The reached final state *)
Definition example1_final_state' :=
 [registers :=
   (update
     (update
       (update
         (update
           (update
             (update
               (update
                 (update
                   (update
                     (update
                       ([r:register]
                         Cases r of
                           (0) => (2)
                         | (S n) => 
                            Cases n of
                              (0) => (3)
                            | (S _) => (0)
                            end
                         end) (1) (2)) (0)
                     (S
                       (update
                         ([r:register]
                           Cases r of
                             (0) => (2)
                           | (S n) => 
                              Cases n of
                                (0) => (3)
                              | (S _) => (0)
                              end
                           end) (1) (2) (0)))) (2) (0)) (1) (1)) (0)
               (S
                 (update
                   (update
                     (update
                       (update
                         ([r:register]
                           Cases r of
                             (0) => (2)
                           | (S n) => 
                              Cases n of
                                (0) => (3)
                              | (S _) => (0)
                              end
                           end) (1) (2)) (0)
                       (S
                         (update
                           ([r:register]
                             Cases r of
                               (0) => (2)
                             | (S n) => 
                                Cases n of
                                  (0) => (3)
                                | (S _) => (0)
                                end
                             end) (1) (2) (0)))) (2) (0)) (1) (1) (0))))
             (2) (0)) (1) (0)) (0)
         (S
           (update
             (update
               (update
                 (update
                   (update
                     (update
                       (update
                         ([r:register]
                           Cases r of
                             (0) => (2)
                           | (S n) => 
                              Cases n of
                                (0) => (3)
                              | (S _) => (0)
                              end
                           end) (1) (2)) (0)
                       (S
                         (update
                           ([r:register]
                             Cases r of
                               (0) => (2)
                             | (S n) => 
                                Cases n of
                                  (0) => (3)
                                | (S _) => (0)
                                end
                             end) (1) (2) (0)))) (2) (0)) (1) (1)) (0)
                 (S
                   (update
                     (update
                       (update
                         (update
                           ([r:register]
                             Cases r of
                               (0) => (2)
                             | (S n) => 
                                Cases n of
                                  (0) => (3)
                                | (S _) => (0)
                                end
                             end) (1) (2)) (0)
                         (S
                           (update
                             ([r:register]
                               Cases r of
                                 (0) => (2)
                               | (S n) => 
                                  Cases n of
                                    (0) => (3)
                                  | (S _) => (0)
                                  end
                               end) (1) (2) (0)))) (2) (0)) (1) (1) (0))))
               (2) (0)) (1) (0) (0)))) (2) (0)) (1) (0))
 ] (Build_configuration (S (S (S O))) registers).

Lemma example1_lemma:
 (RAM_configuration_equality
   (registers_number example1_RAM) example1_final_state example1_final_state').
Proof.
 Split;
  [ Reflexivity
  | Do 3 (* The RAM has 3 registers *)
     (Intro r;
      Case r;
       [ Reflexivity
       | Clear r
       ]
     );
    Reflexivity
  ].
Qed.

Theorem example1 :
 (reduces example1_RAM example1_initial_state example1_final_state').
Proof.
 Split;
  [ Unfold step_star;
    Unfold TransitionSystems.step_star;
    Repeat
     (EApply Rstar_n;
      [ Split;
        [ Simpl;
          Progress Auto
        | Simpl;
          Constructor
        ]
      | Simpl
      ]);
    Constructor 1
  | Unfold stuck;
    Unfold not;
    Intro;
    Elim H;
    Intros;
    Unfold step in H0;
    Elim H0;
    Intros;
    Simpl in H1;
    Generalize (lt_not_le (3) (4));
    Intros;
    Apply H3;
     [ Auto
     | Assumption
     ]
  ].
Qed.
