Require RAMs.
Require Linda.
Require PolyList.
Require TheoryList_compl.
Require Gt.
Require Lt.
Require Le.
Require Plus.
Require Compare_dec.
Require Arith_compl.
Require Peano_dec.
Require RAMs_to_Linda.
Require Minus.
Require Suresnes.BDD.rauzy.algorithme1.Prelude_BDT.
Require MiscRsa.

(*** Inductive predicates that characterize images of RAMs ***)

Inductive is_RAM_instruction [registers_no : nat ; pc : nat] :
 program -> Prop
:=
   IRISucc :
    (r:message)
     (lt r registers_no) ->
      (is_RAM_instruction registers_no pc
       (Action (Out r) (Action (Out (S pc)) Dead)))
 | IRIDecJump :
    (r,n:message)
     (lt r registers_no) ->
     (ge n registers_no) ->
      (is_RAM_instruction registers_no pc
         (IfThenElseAction (Linda.In r)
           (Action (Out (S pc)) Dead)
           (Action (Out n) Dead))).

Section is_RAM_instruction_decidable.

Tactic Definition W := Right; Intro; Inversion H.

Lemma is_RAM_instruction_decidable:
 (registers_no:nat ; pc:nat ; P:program)
  {is_RAM_instruction registers_no pc P} +
  {~(is_RAM_instruction registers_no pc P)}.
Proof.
 Intros;
 Case P;
  [ W 
  | Induction a;
     [ W
     | W
     | Intros;
       Elim (le_lt_dec registers_no m);
        [ W ;
          Elim (le_not_lt ? ? a0 H1) 
        | Intro;
          Case p;
           [ W
           | Intros;
             Case p0;
              [ Case a0;
                 [ W
                 | W
                 | Intro;
                   Elim (eq_nat_dec m0 (S pc));
                    [ Intro;
                      Rewrite a1;
                      Left;
                      Constructor;
                      Assumption
                    | W;
                      Elim b0;
                      Apply sym_eq;
                      Exact H2
                    ]
                 | W
                 ]
              | W
              | W
              | W
              ]
           | W
           | W
           ]
        ]
     | W
     ]
  | Intros;
    Case a;
     [ Intro; 
       Elim (le_lt_dec registers_no m);
        [ W;
          Elim (le_not_lt ? ? a0 H2)
        | Intro;
          Case p;
           [ W
           | Intros;
             Case p1;
              [ Case a0;
                 [ W
                 | W
                 | Intro;
                   Elim (eq_nat_dec (S pc) m0);
                     [ Intro;
                       Rewrite <- a1;
                       Case p0;
                        [ W
                        | Intros;
                          Case p2;
                           [ Case a2;
                              [ W
                              | W
                              | Intro;
                                Elim (le_lt_dec registers_no m1);
                                 [ Intro;
                                   Left;
                                   Constructor;
                                    [ Assumption
                                    | Assumption
                                    ]
                                 | W;
                                   Elim (le_not_lt ? ? H3 b0)
                                 ]
                              | W
                              ]
                           | W
                           | W
                           | W
                           ]
                        | W
                        | W
                        ]
                     | W;
                       Elim (b0 H1)
                     ]
                 | W
                 ]
              | W
              | W
              | W
              ]
           | W
           | W
           ]
        ]
     | W
     | W
     | W
     ]
  | W
  ].
Defined.

End is_RAM_instruction_decidable.


Inductive is_RAM_image_no_pc [registers_no : nat] : agent -> Prop :=
   IRI0Message :
    (m:message)
     (lt m registers_no) ->
      (is_RAM_image_no_pc registers_no (Message m))
 | IRI0AParallel1 :
    (P,Q:agent)
     (is_RAM_image_no_pc registers_no P) ->
      (is_RAM_image_no_pc registers_no Q) ->
       (is_RAM_image_no_pc registers_no (AParallel P Q))
 | IRI0Dead :
    (is_RAM_image_no_pc registers_no (Program Dead))
 | IRI0BangIn :
    (m:message; P:program)
     (ge m registers_no) ->
     (is_RAM_instruction registers_no m P) ->
      (is_RAM_image_no_pc registers_no (Program (Action (BangIn m) P))).

Lemma is_RAM_image_no_pc_rec:
 (registers_no:nat; P:(agent->Set))
  ((m:message)(lt m registers_no)->(P (Message m)))
  ->((P0,Q:agent)
      (is_RAM_image_no_pc registers_no P0)
      ->(P P0)
      ->(is_RAM_image_no_pc registers_no Q)
      ->(P Q)
      ->(P (AParallel P0 Q)))
  ->(P (Program Dead))
  ->((m:message; P0:program)
      (ge m registers_no)
      ->(is_RAM_instruction registers_no m P0)
      ->(P (Program (Action (BangIn m) P0))))
  ->(a:agent)(is_RAM_image_no_pc registers_no a)->(P a).
Proof.
 Induction a;
  [ Intros;
    Elim (le_lt_dec registers_no m);
     [ Intro;
       ElimType False;
       Inversion H3;
       Apply (le_not_lt ? ? a0 H5)
     | Intro;
       Exact (H m b)
     ]
  | Intro p;
    Case p;
     [ Intro;
       Exact H1
     | Intros;
       Induction a0;
        [ ElimType False;
          Inversion H3
        | ElimType False;
          Inversion H3
        | ElimType False;
          Inversion H3
        | Elim (is_RAM_instruction_decidable registers_no m p0);
           [ Intro;
             Apply H2;
              [ Inversion H3;
                Assumption
              | Assumption
              ]
           | Intro;
             ElimType False;
             Inversion H3;
             Exact (b H7)
           ]
        ]
     | Intros;
       ElimType False;
       Inversion H3
     | Intros;
       ElimType False;
       Inversion H3
     ]
  | Intros;
    Cut (is_RAM_image_no_pc registers_no a1);
     [ Intro;
       Cut (is_RAM_image_no_pc registers_no a0);
        [ Intro;
          Apply H0;
           [ Assumption
           | Apply (H3 H7)
           | Assumption
           | Apply (H4 H6)
           ]
        | Inversion H5;
          Assumption
        ]
     | Inversion H5;
       Assumption
     ]
  ].
Qed.

(* The inversion principle in Set *)
Lemma is_RAM_image_no_pc_inv:
 (registers_no:nat; a:agent; P:(nat->agent->Set))
  ((is_RAM_image_no_pc registers_no a)
   ->(m:message)
      (Message m)=a->(lt m registers_no)->(P registers_no a))
  ->((is_RAM_image_no_pc registers_no a)
     ->(P0,Q:agent)
        (AParallel P0 Q)=a
        ->(is_RAM_image_no_pc registers_no P0)
        ->(is_RAM_image_no_pc registers_no Q)
        ->(P registers_no a))
  ->((is_RAM_image_no_pc registers_no a)
     ->(Program Dead)=a
     ->(P registers_no a))
  ->((is_RAM_image_no_pc registers_no a)
     ->(m:message; P0:program)
        (Program (Action (BangIn m) P0))=a
        ->(ge m registers_no)
        ->(is_RAM_instruction registers_no m P0)
        ->(P registers_no a))
  ->(is_RAM_image_no_pc registers_no a)
  ->(P registers_no a).
Proof.
 Intros;
 Induction a;
  [ Apply (H H3 m);
     [ Reflexivity
     | Inversion H3;
       Assumption
     ]
  | Induction p;
     [ Apply (H1 H3);
       Reflexivity
     | Induction a; 
        [ ElimType False;
	  Inversion H3
        | ElimType False;
	  Inversion H3
        | ElimType False;
	  Inversion H3
	| Apply (H2 H3 m p);
	   [ Reflexivity 
	   | Inversion H3;
	     Assumption
	   | Inversion H3;
	     Assumption
	   ]
	]
     | ElimType False;
       Inversion H3
     | ElimType False;
       Inversion H3
     ]
  | Apply (H0 H3 a1 a0);
     [ Reflexivity
     | Inversion H3;
       Assumption
     | Inversion H3;
       Assumption
     ]
  ].
Qed.

(*** Lemmas about RAM_image_no_pc ***)

Lemma RAM_image_no_pc_no_LNot :
 (registers_no:nat ; m:message ; P,P':agent)
  (is_RAM_image_no_pc registers_no P) ->
    ~(OStep P (LNot m) P').
Proof.
 Unfold not;
 Induction P;
  [ Intros;
    Inversion H0
  | Intros;
    Inversion H;
     [ Rewrite <- H2 in H0;
       Inversion H0
     | Rewrite <- H1 in H0;
       Inversion H0
     ]
  | Intros;
    Inversion H1;
    Inversion H2;
     [ Apply (H ? H5 H12)
     | Apply (H0 ? H6 H12)
     | Apply (H ? H5 H10)
     | Apply (H0 ? H6 H10)
     ]
  ].
Qed.

Lemma RAM_image_no_pc_no_LRead :
 (registers_no:nat ; m:message ; P,P':agent)
  (is_RAM_image_no_pc registers_no P) ->
   ~(OStep P (LRead m) P').
Proof.
 Induction P;
  [ Unfold not;
    Intros;
    Inversion H0
  | Inversion 1;
     [ Intro;
       Inversion H0
     | Intro;
       Inversion H3
     ]
  | Unfold not;
    Intros;
    Inversion H1;
    Inversion H2;
     [ Exact (H P'0 H5 H12)
     | Exact (H0 P'0 H6 H12)
     ]
  ].
Qed.

Lemma RAM_image_no_pc_LIn_ge_registers_no :
 (registers_no:nat ; m:message ; P,P':agent)
  (is_RAM_image_no_pc registers_no P) ->
   (OStep P (LIn m) P') ->
    (ge m registers_no).
Proof.
 Induction P;
 [ Intros;
   Inversion H0
 | Induction p;
    [ Intros;
      Inversion H0
    | Intros;
      Inversion H0;
      Inversion H1;
       [ Rewrite <- H2 in H7;
         Discriminate
       | Rewrite <- H2 in H7;
         Rewrite H9 in H7;
         Injection H7;
         Intro;
         Rewrite H10;
         Assumption
       ]
    | Intros;
      Inversion H1
    | Intros;
      Inversion H1
    ]
 | Intros;
   Inversion H1;
   Inversion H2;
    [ Exact (H P'0 H5 H12)
    | Exact (H0 P'0 H6 H12)
    ]
 ].
Qed.
 
Lemma RAM_image_no_pc_LOut_lt_registers_no :
 (registers_no:nat ; m:message ; P,P':agent)
  (is_RAM_image_no_pc registers_no P) ->
   (OStep P (LOut m) P') ->
    (lt m registers_no).
Proof.
 Induction P;
  [ Intros;
    Inversion H;
    Inversion H0;
    Rewrite H5 in H2;
    Assumption
  | Inversion 1;
     [ Intro;
       Inversion H0
     | Intro;
       Inversion H3
     ]
  | Intros;
    Inversion H1;
    Inversion H2;
     [ Exact (H P'0 H5 H12)
     | Exact (H0 P'0 H6 H12)
     ]
  ].
Qed.

Lemma RAM_image_no_pc_LOut_RAM_image_no_pc :
 (registers_no:nat ; m:message ; P,P':agent)
  (is_RAM_image_no_pc registers_no P) ->
   (OStep P (LOut m) P') ->
    (is_RAM_image_no_pc registers_no P').
Proof.
 Intros;
 Generalize P' H0;
 Clear P' H0;
 Induction H;
  [ Intros;
    Inversion H0;
    Constructor
  | Intros;
    Inversion H2;
     [ Constructor;
        [ Apply (HrecH1 ? H7)
	| Assumption
	]
     | Constructor;
        [ Assumption
	| Apply (HrecH0 ? H7)
	]
     ]
  | Intros;
    Inversion H0
  | Intros;
    Inversion H1
  ].
Qed.

Theorem stuck_RAM_image_no_pc :
 (registers_no:nat ; P:agent)
  (is_RAM_image_no_pc registers_no P) -> (stuck P).
Proof.
 Intros registers_no P;
 Unfold stuck TransitionSystems.stuck not;
 Elim P;
  [ Induction 2;
    Intros;
    Inversion H1;
     [ Inversion H2
     | Elim H2;
       Intros;
       Inversion H3
     ]
  | Induction p;
     [ Induction 2;
       Intros;
       Inversion H1;
        [ Inversion H2
        | Inversion H2;
          Inversion H3
        ]
     | Intros;
       Inversion H0;
       Rewrite <- H2 in H1;
       Elim H1;
       Intros;
       Elim H6;
        [ Intro;
          Inversion H7
        | Induction 1;
          Intros;
          Inversion H8
        ]
     | Intros;
       Inversion H1
     | Intros;
       Inversion H1
     ]
  | Intros;
    Inversion H1;
    Elim H2;
    Intros;
    Elim H7;
     [ Intro;
       Inversion H8;
        [ Apply (H H5);
          Exists P';
          Left;
          Assumption
        | Apply (H0 H6);
          Exists P';
          Left;
          Assumption
        | Generalize (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? H5 H11);
          Intro;
          Generalize (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? H6 H12);
          Intro;
          Apply (le_not_lt ? ? H14 H15)
        | Generalize (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? H5 H11);
          Intro;
          Generalize (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? H6 H12);
          Intro;
          Apply (le_not_lt ? ? H15 H14)
        | Apply (RAM_image_no_pc_no_LRead ? ? ? ? H5 H11)
        | Apply (RAM_image_no_pc_no_LRead ? ? ? ? H6 H12)
        ]
     | Induction 1;
       Intros;
       Inversion H9;
        [ Apply (H H5);
          Exists P';
          Right;
          Exists x0;
          Assumption
        | Apply (H0 H6);
          Exists P';
          Right;
          Exists x0;
          Assumption
        | Apply (H H5);
          Exists P';
          Right;
          Exists x0;
          Assumption
        | Apply (H0 H6);
          Exists P';
          Right;
          Exists x0;
          Assumption
        ]
     ]
  ].
Qed.

Inductive is_RAM_image_pc [registers_no : nat] : agent -> Prop :=
   IRIMessage :
    (m:nat)
     (ge m registers_no) ->
      (is_RAM_image_pc registers_no (Message m))
 | IRIAParallel1 :
    (P,Q:agent)
     (is_RAM_image_pc registers_no P) -> (is_RAM_image_no_pc registers_no Q) ->
      (is_RAM_image_pc registers_no (AParallel P Q))
 | IRIAParallel2 :
    (P,Q:agent)
     (is_RAM_image_no_pc registers_no P) -> (is_RAM_image_pc registers_no Q) ->
      (is_RAM_image_pc registers_no (AParallel P Q)).

Lemma is_RAM_image_no_pc_not_is_RAM_image_pc :
 (registers_no:nat ; P:agent)
  (is_RAM_image_no_pc registers_no P) ->
   ~(is_RAM_image_pc registers_no P).
Proof.
 Induction P;
  [ Intros;
    Intro;
    Inversion H;
    Inversion H0;
    Elim (le_not_lt ? ? H4 H2)
  | Intros;
    Intro;
    Inversion H0
  | Intros;
    Intro;
    Inversion H1;
    Inversion H2;
     [ Exact (H H5 H9)
     | Exact (H0 H6 H10)
     ]
  ].
Qed.

Lemma is_RAM_image_no_pc_decidable:
 (registers_no:nat ; P:agent)
  {is_RAM_image_no_pc registers_no P} + {~(is_RAM_image_no_pc registers_no P)}.
Proof.
 Induction P;
  [ Intro;
    Elim (le_lt_dec registers_no m);
     [ Right;
       Intro;
       Inversion H;
       Elim (le_not_lt ? ? a H1)
     | Left;
       Constructor;
       Assumption
     ]
  | Induction p;
     [ Left;
       Constructor
     | Induction a;
        [ Right;
          Intro;
          Inversion H0
        | Right;
          Intro;
          Inversion H0
        | Right;
          Intro;
          Inversion H0
        | Intros;
          Elim (le_lt_dec registers_no m);
           [ Intro;
             Elim (is_RAM_instruction_decidable registers_no m p0);
              [ Left;
                Constructor;
                 [ Exact a0
                 | Assumption
                 ]
              | Right;
                Intro;
                Inversion H0;
                Exact (b H4)
              ]
           | Right;
             Intro;
             Inversion H0;
             Elim (le_not_lt ? ? H3 b)
           ]
        ]
     | Right;
       Intro;
       Inversion H1
     | Right;
       Intro;
       Inversion H1
     ]
  | Intros;
    Elim H;
     [ Intro;
       Elim H0;
        [ Left;
          Constructor;
           [ Assumption
           | Assumption
           ]
        | Right;
          Intro;
          Inversion H1;
          Exact (b H5)
        ]
     | Right;
       Intro;
       Inversion H1;
       Exact (b H4)
     ]
  ].
Defined.

Lemma is_RAM_image_pc_decidable:
 (registers_no:nat ; P:agent)
  {(is_RAM_image_pc registers_no P)}+{~(is_RAM_image_pc registers_no P)}.
Proof.
 Induction P;
  [ Intro;
    Elim (le_lt_dec registers_no m);
     [ Left;
       Constructor;
       Exact a
     | Right;
       Intro;
       Inversion H;
       Elim (le_not_lt ? ? H1 b)
     ]
  | Right;
    Intro;
    Inversion H
  | Intros;
    Inversion H;
     [ Elim (is_RAM_image_no_pc_decidable registers_no a0);
        [ Left;
          Exact (IRIAParallel1 ? ? ? H1 a1)
        | Right;
          Intro;
          Inversion H2;
           [ Exact (b H6)
           | Exact (is_RAM_image_no_pc_not_is_RAM_image_pc ? ? H5 H1)
           ]
        ]
     | Elim (is_RAM_image_no_pc_decidable registers_no a);
        [ Intro;
          Elim H0;
           [ Intro;
             Left;
             Exact (IRIAParallel2 ? ? ? a1 a2)
           | Intro;
             Right;
             Intro;
             Inversion H2;
              [ Exact (H1 H5)
              | Exact (b H6)
              ]
           ]
        | Right;
          Intro;
          Inversion H2;
           [ Exact (H1 H5)
           | Exact (b H5)
           ]
        ]
     ]
  ].
Defined.


(* The inversion principle in Set *)
Lemma is_RAM_image_pc_inv :
 (registers_no:nat; a:agent; P:(nat->agent->Set))
  ((is_RAM_image_pc registers_no a)
   ->(m:nat)
      (Message m)=a->(ge m registers_no)->(P registers_no a))
  ->((is_RAM_image_pc registers_no a)
     ->(P0,Q:agent)
        (AParallel P0 Q)=a
        ->(is_RAM_image_pc registers_no P0)
        ->(is_RAM_image_no_pc registers_no Q)
        ->(P registers_no a))
  ->((is_RAM_image_pc registers_no a)
     ->(P0,Q:agent)
        (AParallel P0 Q)=a
        ->(is_RAM_image_no_pc registers_no P0)
        ->(is_RAM_image_pc registers_no Q)
        ->(P registers_no a))
  ->(is_RAM_image_pc registers_no a)
  ->(P registers_no a).
Proof.
 Intros;
 Induction a;
  [ Apply (H H2 m (refl_equal ? (Message m)));
    Inversion H2;
    Assumption
  | ElimType False;
    Inversion H2
  | Elim (is_RAM_image_pc_decidable registers_no a1);
     [ Intro;
       Cut (is_RAM_image_no_pc registers_no a0);
        [ Intro;
          Apply (H0 H2 ? ? (refl_equal ? (AParallel a1 a0)) a H3)
        | Inversion H2;
           [ Assumption
           | Elim (is_RAM_image_no_pc_not_is_RAM_image_pc ? ? H5 a)
           ]
        ]
     | Intro;
       Cut
        (is_RAM_image_pc registers_no a0)/\(is_RAM_image_no_pc registers_no a1);
        [ Intro;
          Case H3;
          Intros;
          Apply (H1 H2 ? ? (refl_equal ? (AParallel a1 a0)) H5 H4)
        | Inversion H2;
           [ Elim (b H5)
           | Split;
              [ Assumption
              | Assumption
              ]
           ]
        ]
     ]
  ].
Qed.

(*** Lemmas about RAM_image_pc ***)

Lemma RAM_image_pc_no_LRead :
 (registers_no:nat ; m:message ; P,P':agent)
  (is_RAM_image_pc registers_no P) ->
   ~(OStep P (LRead m) P').
Proof.
 Unfold not;
 Induction P;
  [ Intros;
    Inversion H0
  | Intros;
    Inversion H
  | Intros;
    Inversion H1;
     [ Inversion H2;
        [ Apply (H ? H5 H12)
        | Apply (RAM_image_no_pc_no_LRead ? ? ? ? H6 H12)
        ]
     | Inversion H2;
        [ Apply (RAM_image_no_pc_no_LRead ? ? ? ? H5 H12)
        | Apply (H0 ? H6 H12)
        ]
     ]
  ].
Qed.

Lemma RAM_image_pc_no_LNot :
 (registers_no:nat ; P:agent)
  (is_RAM_image_pc registers_no P) ->
    ~(EX m | (EX P' | (OStep P (LNot m) P'))).
Proof.
 Unfold not;
 Induction P;
  [ Intros;
    Elim H0;
    Intros;
    Elim H1;
    Intros;
    Inversion H2
  | Intros;
    Inversion H
  | Intros;
    Inversion H1;
     [ Elim H2;
       Intros;
       Elim H7;
       Intros;
       Inversion H8;
        [ Apply (H H5);
          Exists x;
          Exists P';
          Assumption
        | Apply (RAM_image_no_pc_no_LNot ? ? ? ? H6 H14)
        | Apply (H H5);
          Exists x;
          Exists P';
          Assumption
        | Apply (RAM_image_no_pc_no_LNot ? ? ? ? H6 H12)
        ]
     | Elim H2;
       Intros;
       Elim H7;
       Intros;
       Inversion H8;
        [ Apply (RAM_image_no_pc_no_LNot ? ? ? ? H5 H14)
        | Apply (H0 H6);
          Exists x;
          Exists P';
          Assumption
        | Apply (RAM_image_no_pc_no_LNot ? ? ? ? H5 H12)
        | Apply (H0 H6);
          Exists x;
          Exists P';
          Assumption
        ]
     ]
  ].
Qed.

Lemma RAM_image_pc_LIn_ge_registers_no :
 (registers_no:nat ; m:message ; P,P':agent)
  (is_RAM_image_pc registers_no P) ->
   (OStep P (LIn m) P') ->
    (ge m registers_no).
Proof.
 Induction P;
  [ Intros;
    Inversion H0
  | Intros;
    Inversion H
  | Intros;
    Inversion H1;
     [ Inversion H2;
        [ Apply (H ? H5 H12)
        | Apply (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? H6 H12)
        ]
     | Inversion H2;
        [ Apply (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? H5 H12)
        | Apply (H0 ? H6 H12)
        ]
     ]
  ].
Qed.

Lemma is_RAM_image_pc_AParallel_is_RAM_image_pc_decidable:
 (registers_no:nat ; P,Q:agent)
  (is_RAM_image_pc registers_no (AParallel P Q)) ->
   {(is_RAM_image_pc registers_no P)}+{(is_RAM_image_pc registers_no Q)}.
Proof.
 Intros;
 Elim (is_RAM_image_pc_decidable registers_no P);
  [ Left;
    Assumption
  | Intro;
    Elim (is_RAM_image_pc_decidable registers_no Q);
     [ Right;
       Assumption
     | Intro;
       ElimType False;
       Inversion H;
        [ Elim (b H2)
        | Elim (b0 H3)
        ]
     ]
  ].
Defined.

Definition pc_of_RAM_image_pc :
 (registers_no:nat ; P:agent)
  (is_RAM_image_pc registers_no P) -> message.
 Induction P;
  [ Intros;
    Exact m
  | Intros;
    ElimType False;
    Inversion H
  | Intros;
   Elim (is_RAM_image_pc_decidable registers_no a);
    [ Assumption
    | Intro;
      Elim (is_RAM_image_pc_decidable registers_no a0);
       [ Assumption
       | Intro;
         ElimType False; (* This is the trick that allows me to prove the *)
                         (* proof-irreleance w.r.t. the third argument.   *)
                         (* The reason is that for empty types the        *)
                         (* elimination principle is trivially            *)
                         (* proof-irrelevant.                             *)
         Inversion H1;
          [ Elim (b H4)
          | Elim (b0 H5)
          ]
       ]
    ]
  ].
Defined.

Lemma pc_of_RAM_image_pc_proof_irrelevant :
 (registers_no:nat ; P:agent)
  (P0,P1:(is_RAM_image_pc registers_no P))
   (pc_of_RAM_image_pc registers_no P P0) =
   (pc_of_RAM_image_pc registers_no P P1).
Proof.
 Induction P;
  [ Reflexivity
  | Intros;
    Inversion P0
  | Intros;
    Simpl;
    Elim (is_RAM_image_pc_decidable registers_no a);
     [ Simpl;
       Reflexivity
     | Elim (is_RAM_image_pc_decidable registers_no a0);
        [ Simpl;
          Reflexivity
        | Simpl;
          Intros;
          Cut (P,P':False)(False_rec message P)=(False_rec message P');
           [ Intro PI;
             Apply PI
           | Intros;
             Elim P2
           ]
        ]
     ]
  ].
Qed.

Lemma pc_of_AParallel_equal_pc1 :
 (registers_no:nat ; P,Q:agent)
  (HPQ:(is_RAM_image_pc registers_no (AParallel P Q)))
   (HP:(is_RAM_image_pc registers_no P))
    (pc_of_RAM_image_pc ? ? HPQ) = (pc_of_RAM_image_pc ? ? HP).
Proof.
 Intros;
 Simpl;
 Elim (is_RAM_image_pc_decidable registers_no P);
  [ Simpl;
    Intro;
    Apply pc_of_RAM_image_pc_proof_irrelevant
  | Intro;
    Elim (b HP)
  ].
Qed.

Lemma pc_of_AParallel_equal_pc2 :
 (registers_no:nat ; P,Q:agent)
  (HPQ:(is_RAM_image_pc registers_no (AParallel P Q)))
   (HQ:(is_RAM_image_pc registers_no Q))
    (pc_of_RAM_image_pc ? ? HPQ) = (pc_of_RAM_image_pc ? ? HQ).
Proof.
 Intros;
 Simpl;
 Elim (is_RAM_image_pc_decidable registers_no P);
  [ Intro;
    Inversion HPQ;
     [ Elim (is_RAM_image_no_pc_not_is_RAM_image_pc ? ? H2 HQ)
     | Elim (is_RAM_image_no_pc_not_is_RAM_image_pc ? ? H1 a)
     ]
  | Simpl;
    Elim (is_RAM_image_pc_decidable registers_no Q);
     [ Simpl;
       Intros;
       Apply pc_of_RAM_image_pc_proof_irrelevant
     | Intros;
       Elim (b HQ)
     ]
  ].
Qed.

Lemma pc_of_RAM_image_pc_ge_registers_no:
 (registers_no:nat ; P:agent)
  (H:(is_RAM_image_pc registers_no P))
   (ge (pc_of_RAM_image_pc ? ? H) registers_no).
Proof.
 Induction P;
  [ Intros;
    Inversion H;
    Simpl;
    Assumption
  | Intros;
    Inversion H
  | Intros;
    Simpl;
    Elim (is_RAM_image_pc_decidable registers_no a);
     [ Simpl;
       Assumption
     | Simpl;
       Intro;
       Elim (is_RAM_image_pc_decidable registers_no a0);
        [ Simpl;
          Assumption
        | Intro;
          Inversion H1;
           [ Elim (b H4)
           | Elim (b0 H5)
           ]
        ]
     ]
  ].
Qed.

Lemma RAM_image_pc_LOut :
 (registers_no:nat ; P:agent)
  (H:(is_RAM_image_pc registers_no P))
   [pc := (pc_of_RAM_image_pc ? ? H)]
    {P':agent | (OStep P (LOut pc) P')}.
Proof.
 Induction P;
  [ Intros;
    Simpl in pc;
    EApply exist;
    Constructor
  | Intros;
    ElimType False;
    Inversion H
  | Intros;
    Inversion H1 using is_RAM_image_pc_inv;
     [ Intros;
       Discriminate H3
     | Intros;
       Injection H3;
       Intros;
       Rewrite H7 in H4;
       Elim (H H4);
       Intros;
       Rewrite <- (pc_of_AParallel_equal_pc1 ? ? ? H1 H4) in p;
       Fold pc in p;
       Exists (AParallel x a0);
       Constructor;
        [ Unfold not;
          Induction 1;
          Intros;
          Discriminate
        | Assumption
        ]
     | Intros;
       Injection H3;
       Intros;
       Rewrite H6 in H5;
       Elim (H0 H5);
       Intros;
       Rewrite <- (pc_of_AParallel_equal_pc2 ? ? ? H1 H5) in p;
       Fold pc in p;
       Exists (AParallel a x);
       Constructor;
        [ Unfold not;
          Induction 1;
          Intros;
          Discriminate
        | Assumption
        ]
     ]
  ].
Defined.

Lemma RAM_image_pc_LOut_RAM_image_no_pc :
 (registers_no:nat ; pc:nat; P,P':agent)
  (ge pc registers_no) ->
   (H0:(is_RAM_image_pc registers_no P))
    (OStep P (LOut pc) P') ->
     (is_RAM_image_no_pc registers_no P').
Proof.
 Intros;
 Generalize P' H1;
 Clear P' H1;
 Induction H0;
  [ Intros;
    Inversion H1;
    Constructor
  | Intros;
    Inversion H2;
     [ Constructor;
        [ Apply (HrecH0 ? H8)
	| Assumption
	]
     | Generalize (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? H1 H8);
       Intro Hlt;
       Elim (le_not_lt ? ? H Hlt)
     ]
  | Intros;
    Inversion H2;
     [ Generalize (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? H0 H8);
       Intro Hlt;
       Elim (le_not_lt ? ? H Hlt)
     | Constructor;
	[ Assumption
        | Apply (HrecH0 ? H8)
	]
     ]
  ].
Qed.

Lemma RAM_image_pc_LOut_deterministic :
 (registers_no:nat ; pc,pc':nat)
  (ge pc registers_no) -> (ge pc' registers_no) ->
   (P,P',P'':agent)
    (is_RAM_image_pc registers_no P) ->
     (OStep P (LOut pc) P') ->
      (OStep P (LOut pc') P'') ->
       pc=pc' /\ P'=P''.
Proof.
 Induction P;
  [ Intros;
    Inversion H2;
    Inversion H3;
    Split;
     [ Rewrite H6 in H9;
       Assumption
     | Reflexivity
     ]
  | Intros;
    Inversion H1
  | Intros;
    Inversion H3;
     [ Inversion H4;
        [ Inversion H5;
           [ Elim (H1 ? ? H8 H15 H21);
             Intros;
             Split;
              [ Assumption
              | Rewrite H23;
                Reflexivity
              ]
           | Generalize (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? H9 H21);
             Intro;
             Elim (le_not_lt ? ? H0 H22)
           ]
        | Inversion H5;
           [ Generalize (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? H9 H15);
             Intro;
             Elim (le_not_lt ? ? H H22)
           | Generalize (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? H9 H21);
             Intro;
             Elim (le_not_lt ? ? H0 H22)
           ]
        ]
     | Inversion H4;
        [ Inversion H5;
           [ Generalize (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? H8 H21);
             Intro;
             Elim (le_not_lt ? ? H0 H22)
           | Generalize (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? H8 H15);
             Intro;
             Elim (le_not_lt ? ? H H22)
           ]
        | Inversion H5;
           [ Generalize (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? H8 H21);
             Intro;
             Elim (le_not_lt ? ? H0 H22)
           | Elim (H2 ? ? H9 H15 H21);
             Intros;
             Split;
              [ Assumption
              | Rewrite H23;
                Reflexivity
              ]
           ]
        ]
     ]
  ].
Qed.

Lemma RAM_image_pc_LOut_pc :
 (registers_no:nat ; pc:nat ; P,P':agent)
  (ge pc registers_no) ->
    (H:(is_RAM_image_pc registers_no P))
     (OStep P (LOut pc) P') ->
       pc=(pc_of_RAM_image_pc ? ? H).
Proof.
 Intros;
 Elim (RAM_image_pc_LOut ? ? H0);
 Intros;
 Generalize (pc_of_RAM_image_pc_ge_registers_no ? ? H0);
 Intro;
 Elim (RAM_image_pc_LOut_deterministic ? ? ? H H2 ? ? ? H0 H1 p);
 Intros;
 Assumption.
Qed.


Inductive is_RAM_image_succ [registers_no:nat ; r:register ; pc:nat] :
 agent -> Prop
:=
   IRISOutSucc :
    (is_RAM_image_succ registers_no r pc
     (Program (Action (Out r) (Action (Out (S pc)) Dead))))
 | IRISAParallel1 :
    (P,Q:agent)
     (is_RAM_image_succ registers_no r pc P) ->
      (is_RAM_image_no_pc registers_no Q) ->
       (is_RAM_image_succ registers_no r pc (AParallel P Q))
 | IRISAParallel2 :
    (P,Q:agent)
     (is_RAM_image_no_pc registers_no P) ->
      (is_RAM_image_succ registers_no r pc Q) ->
       (is_RAM_image_succ registers_no r pc (AParallel P Q)).

Lemma is_RAM_image_no_pc_not_is_RAM_image_succ:
 (registers_no:nat ; r:register ; pc:nat ; a:agent)
  (is_RAM_image_no_pc registers_no a) ->
   ~(is_RAM_image_succ registers_no r pc a).
Proof.
 Induction 1;
  [ Intros;
    Intro;
    Inversion H1
  | Intros;
    Intro;
    Inversion H4;
     [ Apply (H1 H7)
     | Apply (H3 H8)
     ]
  | Intro;
    Inversion H0
  | Intros;
    Intro;
    Inversion H2
  ].
Qed.

Lemma is_RAM_image_succ_decidable:
 (registers_no:nat ; r:register ; pc:nat ; a:agent)
  {(is_RAM_image_succ registers_no r pc a)} +
   {~(is_RAM_image_succ registers_no r pc a)}.
Proof.
 Induction a;
  [ Right;
    Intro;
    Inversion H
  | Induction p;
     [ Right;
       Intro;
       Inversion H
     | Intros;
       Induction a0;
        [ Right;
          Intro;
          Inversion H0
        | Right;
          Intro;
          Inversion H0
        | Induction p0;
           [ Right;
             Intro;
             Inversion H0
           | Induction a0;
              [ Right;
                Intro;
                Inversion H0
              | Right;
                Intro;
                Inversion H0
              | Induction p0;
                 [ Elim (eq_nat_dec m r);
                    [ Intro;
                      Rewrite a0;
                      Elim (eq_nat_dec m0 (S pc));
                       [ Left;
                         Rewrite a1;
                         Constructor
                       | Right;
                         Intro;
                         Inversion H0;
                         Apply b;
                         Symmetry;
                         Assumption
                       ]
                    | Right;
                      Intro;
                      Inversion H0;
                      Apply b;
                      Symmetry;
                      Assumption
                    ]
                 | Right;
                   Intro;
                   Inversion H0
                 | Right;
                   Intro;
                   Inversion H0
                 | Right;
                   Intro;
                   Inversion H0
                 ]
              | Right;
                Intro;
                Inversion H0
              ]
           | Right;
             Intro;
             Inversion H0
           | Right;
             Intro;
             Inversion H0
           ]
        | Right;
          Intro;
          Inversion H0
        ]
     | Right;
       Intro;
       Inversion H1
     | Right;
       Intro;
       Inversion H1
     ]
  | Intros;
    Elim H;
     [ Intro;
       Elim (is_RAM_image_no_pc_decidable registers_no a1);
        [ Left;
          Apply IRISAParallel1;
           [ Assumption
           | Assumption
           ]
        | Right;
          Intro;
          Inversion H1;
           [ Apply (b H5)
           | Apply (is_RAM_image_no_pc_not_is_RAM_image_succ ? ? ? ? H4 a2)
           ]
        ]
     | Intro;
       Elim H0;
        [ Intro;
          Elim (is_RAM_image_no_pc_decidable registers_no a0);
           [ Left;
             Apply IRISAParallel2;
              [ Assumption
              | Assumption
              ]
           | Right;
             Intro;
             Inversion H1;
              [ Apply (is_RAM_image_no_pc_not_is_RAM_image_succ ? ? ? ? H5 a2)
              | Apply (b0 H4)
              ]
           ]
        | Right;
          Intro;
          Inversion H1;
           [ Apply (b H4)
           | Apply (b0 H5)
           ]
        ]
     ]
  ].
Qed.

Lemma is_RAM_image_succ_rec :
 (registers_no:nat; r:register; pc:nat; P:(agent->Set))
  (P (Program (Action (Out r) (Action (Out (S pc)) Dead))))
  ->((P0,Q:agent)
      (is_RAM_image_succ registers_no r pc P0)
      ->(P P0)
      ->(is_RAM_image_no_pc registers_no Q)
      ->(P (AParallel P0 Q)))
  ->((P0,Q:agent)
      (is_RAM_image_no_pc registers_no P0)
      ->(is_RAM_image_succ registers_no r pc Q)
      ->(P Q)
      ->(P (AParallel P0 Q)))
  ->(a:agent)(is_RAM_image_succ registers_no r pc a)->(P a).
Proof.
 Induction a;
  [ Intros;
    ElimType False;
    Inversion H2
  | Intros;
    Induction p;
     [ ElimType False;
       Inversion H2
     | Induction a0;
        [ ElimType False;
          Inversion H2
        | ElimType False;
          Inversion H2
        | Induction p;
           [ ElimType False;
             Inversion H2
           | Induction a0;
              [ ElimType False;
                Inversion H2
              | ElimType False;
                Inversion H2
              | Induction p;
                 [ Elim (eq_nat_dec m0 (S pc));
                    [ Intro;
                      Rewrite a0;
                      Cut r=m;
                       [ Intro;
                         Rewrite <- H3;
                         Assumption
                       | Inversion H2;
                         Reflexivity
                       ]
                    | Intro;
                      Elim b;
                      Inversion H2;
                      Reflexivity
                    ]
                 | ElimType False;
                   Inversion H2
                 | ElimType False;
                   Inversion H2
                 | ElimType False;
                   Inversion H2
                 ]
              | ElimType False;
                Inversion H2
              ]
           | ElimType False;
             Inversion H2
           | ElimType False;
             Inversion H2
           ]
        | ElimType False;
          Inversion H2
        ]
     | ElimType False;
       Inversion H2
     | ElimType False;
       Inversion H2
     ]
  | Intros;
    Elim (is_RAM_image_succ_decidable registers_no r pc a0);
     [ Intro;
       Cut (is_RAM_image_no_pc registers_no a1);
        [ Intro;
          Apply H0;
           [ Assumption
           | Apply (H2 a2)
           | Assumption
           ]
        | Inversion H4;
           [ Assumption
           | Elim
              (is_RAM_image_no_pc_not_is_RAM_image_succ ? ? ? ? H7 a2)
           ]
        ]
     | Intro;
       Cut
        (is_RAM_image_succ registers_no r pc a1) /\
        (is_RAM_image_no_pc registers_no a0);
        [ Intro;
          Elim H5;
          Intros;
          Apply H1;
           [ Assumption
           | Assumption
           | Apply (H3 H6) 
           ]
        | Inversion H4;
           [ Elim (b H7)
           | Split;
              [ Assumption
              | Assumption
              ]
           ]
        ]
     ]
  ].
Qed.

Inductive is_RAM_image_decjump
 [registers_no:nat ; r:register ; pc':nat ; pc:nat] : agent -> Prop
:=
   IRIDCOutDecJump :
    (is_RAM_image_decjump registers_no r pc' pc
     (Program 
      (IfThenElseAction (Linda.In r)
        (Action (Out (S pc)) Dead)
        (Action (Out pc') Dead))))
 | IRIDCAParallel1 :
    (P,Q:agent)
     (is_RAM_image_decjump registers_no r pc' pc P) ->
      (is_RAM_image_no_pc registers_no Q) ->
       (is_RAM_image_decjump registers_no r pc' pc (AParallel P Q))
 | IRIDCAParallel2 :
    (P,Q:agent)
     (is_RAM_image_no_pc registers_no P) ->
      (is_RAM_image_decjump registers_no r pc' pc Q) ->
       (is_RAM_image_decjump registers_no r pc' pc (AParallel P Q)).

Lemma RAM_image_decjump_LIn_eq_or_ge_registers_no :
 (registers_no:nat ; r,m:message ; pc,pc':nat ; P,P':agent)
  (is_RAM_image_decjump registers_no r pc' pc P) ->
   (OStep P (LIn m) P') ->
    m=r\/(ge m registers_no).
Proof.
 Induction P;
 [ Intros;
   Inversion H0
 | Induction p;
    [ Intros;
      Inversion H0
    | Intros;
      Inversion H0
    | Intros;
      Inversion H2;
      Inversion H1;
      Left;
      Rewrite <- H7;
      Rewrite <- H9 in H4;
      Injection H4;
      Intro;
      Assumption
    | Intros;
      Inversion H1
    ]
 | Intros;
   Inversion H1;
    [ Inversion H2;
       [ Apply (H ? H5 H12)
       | Right;
         Apply (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? H6 H12)
       ]
    | Inversion H2;
       [ Right;
         Apply (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? H5 H12)
       | Apply (H0 ? H6 H12)
       ]
    ]
 ].
Qed.

Lemma RAM_image_decjump_LOut_lt_registers_no:
 (registers_no:nat ; m,m':message ; pc,pc':nat ; P,P':agent)
  (is_RAM_image_decjump registers_no m pc pc' P) ->
   (OStep P (LOut m') P') ->
    (lt m' registers_no).
Proof.
 Induction P;
  [ Intros;
    Inversion H
  | Intros;
    Inversion H;
    Rewrite <- H2 in H0;
    Inversion H0
  | Intros;
    Inversion H1;
     [ Inversion H2;
        [ Exact (H ? H5 H12)
        | Apply (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? H6 H12)
        ]
     | Inversion H2;
        [ Apply (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? H5 H12)
        | Exact (H0 ? H6 H12)
        ]
     ]
  ].
Qed.

Lemma RAM_image_decjump_not_LRead:
 (registers_no:nat ; m,m':message ; pc,pc':nat ; P,P':agent)
  (is_RAM_image_decjump registers_no m pc pc' P) ->
   ~(OStep P (LRead m') P').
Proof.
 Induction P;
  [ Intros;
    Inversion H
  | Intros;
    Intro;
    Inversion H;
    Rewrite <- H2 in H0;
    Inversion H0
  | Intros;
    Intro;
    Inversion H1;
     [ Inversion H2;
        [ Exact (H ? H5 H12)
        | Elim (RAM_image_no_pc_no_LRead ? ? ? ? H6 H12)
        ]
     | Inversion H2;
        [ Elim (RAM_image_no_pc_no_LRead ? ? ? ? H5 H12)
        | Exact (H0 ? H6 H12)
        ]
     ]
  ].
Qed.


(* Technical induction principle *)
Lemma is_RAM_image_decjump_Tau_ind:
 (registers_no:nat; r:register; pc',pc:nat; P:(agent->agent->Prop))
  ((P0,P',Q:agent)(P P0 P') -> (P (AParallel P0 Q) (AParallel P' Q))) ->
  ((P0,Q,Q':agent)(P Q Q') -> (P (AParallel P0 Q) (AParallel P0 Q'))) ->
  ((P0,P',Q,Q':agent)
    (is_RAM_image_decjump registers_no r pc' pc P0) ->
     (is_RAM_image_no_pc registers_no Q) ->
      (OStep P0 (LIn r) P') ->
       (OStep Q (LOut r) Q') ->
        (P (AParallel P0 Q) (AParallel P' Q'))) ->
  ((P0,P',Q,Q':agent)
    (is_RAM_image_no_pc registers_no P0) ->
     (is_RAM_image_decjump registers_no r pc' pc Q) ->
      (OStep P0 (LOut r) P') ->
       (OStep Q (LIn r) Q') ->
        (P (AParallel P0 Q) (AParallel P' Q'))) ->
  (a,a':agent)
   (is_RAM_image_decjump registers_no r pc' pc a) ->
    (OStep a Tau a') ->
     (P a a').
Proof.
 Intros;
 Generalize a' H4;
 Clear a' H4;
 Induction H3;
  [ Intros;
    Inversion H4
  | Intros;
    Inversion H5;
     [ Apply H;
       Exact (HrecH3 ? H11)
     | Elim (stuck_RAM_image_no_pc ? ? H4);
       Exists P';
       Left;
       Assumption
     | Elim
        (RAM_image_decjump_LIn_eq_or_ge_registers_no ? ? ? ? ? ? ? H3 H8);
	[ Intro;
	  Rewrite H11 in H8;
	  Rewrite H11 in H9;
	  Apply (H1 P0 P' Q Q' H3 H4);
	   [ Assumption
	   | Assumption
	   ]
	| Intro Hge;
          Generalize
           (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? H4 H9);
          Intro Hlt;
	  Elim (le_not_lt ? ? Hge Hlt)
	]
     | Generalize
        (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? H4 H9);
       Intro Hge;
       Generalize
        (RAM_image_decjump_LOut_lt_registers_no ? ? ? ? ? ? ? H3 H8);
       Intro Hlt;
       Elim (le_not_lt ? ? Hge Hlt)
     | Elim
        (RAM_image_decjump_not_LRead ? ? ? ? ? ? ? H3 H8)
     | Elim (RAM_image_no_pc_no_LRead ? ? ? ? H4 H9)
     ]
  | Intros;
    Inversion H5;
     [ Elim (stuck_RAM_image_no_pc ? ? H3);
       Exists P';
       Left;
       Assumption
     | Apply H0;
       Exact (HrecH3 ? H11)
     | Generalize
        (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? H3 H8);
       Intro Hge;
       Generalize
        (RAM_image_decjump_LOut_lt_registers_no ? ? ? ? ? ? ? H4 H9);
       Intro Hlt;
       Elim (le_not_lt ? ? Hge Hlt)
     | Elim
        (RAM_image_decjump_LIn_eq_or_ge_registers_no ? ? ? ? ? ? ? H4 H9);
	[ Intro;
	  Rewrite H11 in H8;
	  Rewrite H11 in H9;
	  Apply (H2 P0 P' Q Q' H3 H4);
	   [ Assumption
	   | Assumption
	   ]
	| Intro Hge;
          Generalize
           (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? H3 H8);
          Intro Hlt;
	  Elim (le_not_lt ? ? Hge Hlt)
	]
     | Elim (RAM_image_no_pc_no_LRead ? ? ? ? H3 H8)
     | Elim
        (RAM_image_decjump_not_LRead ? ? ? ? ? ? ? H4 H9)
     ]
  ].
Qed.


Lemma is_RAM_image_no_pc_not_is_RAM_image_decjump:
 (registers_no:nat; r:register; pc',pc:nat; a:agent)
  (is_RAM_image_no_pc registers_no a) ->
   ~(is_RAM_image_decjump registers_no r pc' pc a).
Proof.
 Intros;
 Intro;
 Induction H;
  [ Inversion H0
  | Inversion H0;
     [ Exact (HrecH1 H4)
     | Exact (HrecH0 H5)
     ]
  | Inversion H0
  | Inversion H0
  ].
Qed.

Lemma is_RAM_image_decjump_decidable:
 (registers_no:nat; r:register; pc',pc:nat; a:agent) 
  {(is_RAM_image_decjump registers_no r pc' pc a)} +
   {~(is_RAM_image_decjump registers_no r pc' pc a)}.
Proof.
 Induction a;
  [ Right;
    Intro;
    Inversion H
  | Induction p;
     [ Right;
       Intro;
       Inversion H
     | Right;
       Intro;
       Inversion H0
     | Intros;
       Induction a0;
        [ Induction p0;
	   [ Right;
             Intro;
             Inversion H1
	   | Induction p0;
	      [ Induction p1;
                 [ Right;
                   Intro;
                   Inversion H1
		 | Induction p1;
		    [ Induction a0;
                       [ Right;
                         Intro;
                         Inversion H1
                       | Right;
                         Intro;
                         Inversion H1
                       | Induction a1;
                          [ Right;
                            Intro;
                            Inversion H1
                          | Right;
                            Intro;
                            Inversion H1
                          | Elim (eq_nat_dec r m);
			     [ Intro;
			       Rewrite <- a0;
                               Elim (eq_nat_dec (S pc) m0);
			        [ Intro;
				  Rewrite <- a1;
                                  Elim (eq_nat_dec pc' m1);
				   [ Intro;
				     Rewrite <- a2;
				     Left;
				     Constructor
				   | Right;
				     Intro;
				     Inversion H1;
				     Exact (b H3)
				   ]
				| Right;
				  Intro;
				  Inversion H1;
				  Exact (b H3)
				]
			     | Intro;
			       Right;
			       Intro;
			       Inversion H1;
			       Exact (b H3)
			     ]
                          | Right;
                            Intro;
                            Inversion H1
                          ]
                       | Right;
                         Intro;
                         Inversion H1
                       ]
                    | Right;
                      Intro;
                      Inversion H1
                    | Right;
                      Intro;
                      Inversion H1
                    | Right;
                      Intro;
                      Inversion H1
		    ]
                 | Right;
                   Intro;
                   Inversion H1
                 | Right;
                   Intro;
                   Inversion H1
		 ]
              | Right;
                Intro;
                Inversion H1
              | Right;
                Intro;
                Inversion H1
              | Right;
                Intro;
                Inversion H1
              ]
           | Right;
             Intro;
             Inversion H1
           | Right;
             Intro;
             Inversion H1
	   ]
        | Right;
          Intro;
          Inversion H1
        | Right;
          Intro;
          Inversion H1
        | Right;
          Intro;
          Inversion H1
	]
     | Right;
       Intro;
       Inversion H1
     ]
  | Intros;
    Elim H;
     [ Intro;
       Elim (is_RAM_image_no_pc_decidable registers_no a1);
        [ Intro;
	  Left;
	  Apply IRIDCAParallel1;
	   [ Assumption
	   | Assumption
	   ]
	| Intro;
	  Right;
	  Intro;
	  Inversion H1;
	   [ Exact (b H5)
	   | Elim (is_RAM_image_no_pc_not_is_RAM_image_decjump ? ? ? ? ? H4 a2)
	   ]
	]
     | Intro;
       Elim H0;
        [ Intro;
	  Elim (is_RAM_image_no_pc_decidable registers_no a0);
	   [ Intro;
	     Left;
	     Apply IRIDCAParallel2;
	      [ Assumption
	      | Assumption
	      ]
	   | Intro;
	     Right;
	     Intro;
	     Inversion H1;
	      [ Apply (b H4)
	      | Apply (b0 H4)
	      ]
	   ]
	| Intro;
	  Right;
	  Intro;
	  Inversion H1;
	   [ Exact (b H4)
	   | Exact (b0 H5)
	   ]
	]
     ]
  ].
Qed.

Lemma is_RAM_image_decjump_rec:
 (registers_no:nat; r:register; pc',pc:nat; P:(agent->Set))
  (P
    (Program
      (IfThenElseAction (Linda.In r) (Action (Out (S pc)) Dead)
        (Action (Out pc') Dead))))
  ->((P0,Q:agent)
      (is_RAM_image_decjump registers_no r pc' pc P0)
      ->(P P0)
      ->(is_RAM_image_no_pc registers_no Q)
      ->(P (AParallel P0 Q)))
  ->((P0,Q:agent)
      (is_RAM_image_no_pc registers_no P0)
      ->(is_RAM_image_decjump registers_no r pc' pc Q)
      ->(P Q)
      ->(P (AParallel P0 Q)))
  ->(a:agent)
     (is_RAM_image_decjump registers_no r pc' pc a)->(P a).
Proof.
 Induction a;
  [ Intros;
    ElimType False;
    Inversion H2
  | Induction p;
     [ Intros;
       ElimType False;
       Inversion H2
     | Intros;
       ElimType False;
       Inversion H3
     | Intros;
       Induction a0;
        [ Induction p0;
           [ Intros;
             ElimType False;
             Inversion H4
           | Induction p0;
              [ Induction p1;
                 [ Intros;
                   ElimType False;
                   Inversion H4
                 | Induction p1;
                    [ Induction a0;
                       [ Intros;
                         ElimType False;
                         Inversion H4
                       | Intros;
                         ElimType False;
                         Inversion H4
                       | Induction a1;
                          [ Intros;
                            ElimType False;
                            Inversion H4
                          | Intros;
                            ElimType False;
                            Inversion H4
                          | Elim (eq_nat_dec m r);
                             [ Intro;
                               Rewrite a0;
                               Elim (eq_nat_dec (S pc) m0);
                                [ Intro;
                                  Rewrite <- a1;
                                  Elim (eq_nat_dec pc' m1);
                                   [ Intro;
				     Rewrite <- a2;
				     Assumption
                                   | Intro;
                                     ElimType False;
				     Inversion H4;
				     Apply (b H8)
                                   ]
                                | Intro;
                                  ElimType False;
                                  Inversion H4;
                                  Apply (b H7)
                                ]
                             | Intro;
                               ElimType False;
                               Inversion H4;
                               Apply b;
                               Symmetry;
                               Assumption
                             ]
                          | Intros;
                            ElimType False;
                            Inversion H4
                          ]
                       | Intros;
                         ElimType False;
                         Inversion H4
                       ]
                    | Intros;
                      ElimType False;
                      Inversion H4
                    | Intros;
                      ElimType False;
                      Inversion H4
                    | Intros;
                      ElimType False;
                      Inversion H4
                    ]
                 | Intros;
                   ElimType False;
                   Inversion H4
                 | Intros;
                   ElimType False;
                   Inversion H4
                 ]
              | Intros;
                ElimType False;
                Inversion H4
              | Intros;
                ElimType False;
                Inversion H4
              | Intros;
                ElimType False;
                Inversion H4
              ]
           | Intros;
             ElimType False;
             Inversion H4
           | Intros;
             ElimType False;
             Inversion H4
           ]
        | Intros;
          ElimType False;
          Inversion H4
        | Intros;
          ElimType False;
          Inversion H4
        | Intros;
          ElimType False;
          Inversion H4
        ]
     | Intros;
       ElimType False;
       Inversion H4
     ]
  | Intros;
    Elim (is_RAM_image_decjump_decidable registers_no r pc' pc a0);
     [ Intro;
       Elim (is_RAM_image_no_pc_decidable registers_no a1);
        [ Intro;
	  Apply H0;
	   [ Assumption
	   | Exact (H2 a2)
	   | Assumption
	   ]
	| Intro;
	  ElimType False;
	  Inversion H4;
	   [ Exact (b H8)
	   | Apply (is_RAM_image_no_pc_not_is_RAM_image_decjump ? ? ? ? ? H7 a2)
	   ]
	]
     | Intro;
       Elim (is_RAM_image_decjump_decidable registers_no r pc' pc a1);
        [ Intro;
	  Cut (is_RAM_image_no_pc registers_no a0);
	   [ Intro;
	     Apply H1;
	      [ Assumption
	      | Assumption
	      | Exact (H3 a2)
	      ]
	   | Inversion H4;
	      [ Elim (b H7)
	      | Assumption
	      ]
	   ]
	| Intro;
	  ElimType False;
	  Inversion H4;
	   [ Exact (b H7)
	   | Exact (b0 H8)
	   ]
	]
     ]
  ].
Qed.

Inductive is_RAM_image_out_pc [registers_no:nat ; pc:nat] :
 agent -> Prop
:=
   IRIOPCOutSucc :
    (is_RAM_image_out_pc registers_no pc (Program (Action (Out pc) Dead)))
 | IRIOPCAParallel1 :
    (P,Q:agent)
     (is_RAM_image_out_pc registers_no pc P) ->
      (is_RAM_image_no_pc registers_no Q) ->
       (is_RAM_image_out_pc registers_no pc (AParallel P Q))
 | IRIOPCAParallel2 :
    (P,Q:agent)
     (is_RAM_image_no_pc registers_no P) ->
      (is_RAM_image_out_pc registers_no pc Q) ->
       (is_RAM_image_out_pc registers_no pc (AParallel P Q)).

Lemma is_RAM_image_no_pc_not_is_RAM_image_out_pc:
 (registers_no:nat ; pc:nat ; a:agent)
  (is_RAM_image_no_pc registers_no a) ->
   ~(is_RAM_image_out_pc registers_no pc a).
Proof.
 Induction a;
  [ Intros;
    Intro;
    Inversion H0
  | Intros;
    Inversion H;
     [ Intro;
       Inversion H0
     | Intro;
       Inversion H3
     ]
  | Intros;
    Intro;
    Inversion H1;
    Inversion H2;
     [ Apply (H H5 H9)
     | Apply (H0 H6 H10)
     ]
  ].
Qed.

Lemma is_RAM_image_out_pc_decidable:
 (registers_no:nat ; pc:nat ; a:agent)
  {(is_RAM_image_out_pc registers_no pc a)} +
   {~(is_RAM_image_out_pc registers_no pc a)}.
Proof.
 Induction a;
  [ Right;
    Intro;
    Inversion H
  | Induction p;
     [ Right;
       Intro;
       Inversion H
     | Intros;
       Induction a0;
        [ Right;
          Intro;
          Inversion H0
        | Right;
          Intro;
          Inversion H0
        | Induction p0;
           [ Elim (eq_nat_dec m pc);
              [ Left;
                Rewrite a0;
                Constructor
              | Right;
                Intro;
                Inversion H0;
                Apply b;
                Symmetry;
                Assumption
              ]
           | Right;
             Intro;
             Inversion H0
           | Right;
             Intro;
             Inversion H0
           | Right;
             Intro;
             Inversion H0
           ]
        | Right;
          Intro;
          Inversion H0
        ]
     | Right;
       Intro;
       Inversion H1
     | Right;
       Intro;
       Inversion H1
     ]
  | Intros;
    Elim H;
     [ Intro;
       Elim (is_RAM_image_no_pc_decidable registers_no a1);
        [ Left;
          Apply IRIOPCAParallel1;
           [ Assumption
           | Assumption
           ]
        | Right;
          Intro;
          Inversion H1;
           [ Apply (b H5)
           | Apply (is_RAM_image_no_pc_not_is_RAM_image_out_pc ? ? ? H4 a2)
           ]
        ]
     | Intro;
       Elim H0;
        [ Intro;
          Elim (is_RAM_image_no_pc_decidable registers_no a0);
           [ Left;
             Apply IRIOPCAParallel2;
              [ Assumption
              | Assumption
              ]
           | Right;
             Intro;
             Inversion H1;
              [ Apply (b H4)
              | Apply (b0 H4)
              ]
           ]
        | Right;
          Intro;
          Inversion H1;
           [ Apply (b H4)
           | Apply (b0 H5)
           ]
        ]
     ]
  ].
Qed.

Lemma is_RAM_image_out_pc_rec:
 (registers_no,pc:nat; P:(agent->Set))
  (P (Program (Action (Out pc) Dead)))
  ->((P0,Q:agent)
      (is_RAM_image_out_pc registers_no pc P0)
      ->(P P0)
      ->(is_RAM_image_no_pc registers_no Q)
      ->(P (AParallel P0 Q)))
  ->((P0,Q:agent)
      (is_RAM_image_no_pc registers_no P0)
      ->(is_RAM_image_out_pc registers_no pc Q)
      ->(P Q)
      ->(P (AParallel P0 Q)))
  ->(a:agent)(is_RAM_image_out_pc registers_no pc a)->(P a).
Proof.
 Induction a;
  [ Intros;
    ElimType False;
    Inversion H2
  | Intros;
    Induction p;
     [ ElimType False;
       Inversion H2
     | Induction a0;
        [ ElimType False;
          Inversion H2
        | ElimType False;
          Inversion H2
        | Induction p;
           [ Elim (eq_nat_dec m pc);
              [ Intro;
                Rewrite a0;
                Assumption
              | Intro ;
                ElimType False;
                Inversion H2;
                Apply b;
                Symmetry;
                Assumption
              ]
           | ElimType False;
             Inversion H2
           | ElimType False;
             Inversion H2
           | ElimType False;
             Inversion H2
           ]
        | ElimType False;
          Inversion H2
        ]
     | ElimType False;
       Inversion H2
     | ElimType False;
       Inversion H2
     ]
  | Intros;
    Elim (is_RAM_image_out_pc_decidable registers_no pc a0);
     [ Intro;
       Cut (is_RAM_image_no_pc registers_no a1);
        [ Intro;
          Apply H0;
           [ Assumption
           | Apply H2;
             Assumption
           | Assumption
           ]
        | Inversion H4;
           [ Assumption
           | Elim (is_RAM_image_no_pc_not_is_RAM_image_out_pc ? ? ? H7 a2)
           ]
        ]
     | Intro;
       Cut
        (is_RAM_image_no_pc registers_no a0) /\
         (is_RAM_image_out_pc registers_no pc a1);
        [ Intros;
          Elim H5;
          Intros;
          Apply H1;
           [ Assumption
           | Assumption
           | Apply H3;
             Assumption
           ]
        | Inversion H4;
           [ Elim (b H7)
           | Split;
              [ Assumption
              | Assumption
              ]
           ]
        ]
     ]
  ].
Qed.

Inductive instruction_does_not_occur [registers_no:nat; n:nat] : agent -> Prop
:=
   IOO0Message :
    (m:message)(instruction_does_not_occur registers_no n (Message m))
 | IOO0Dead :
    (instruction_does_not_occur registers_no n (Program Dead))
 | IOO0BangIn :
    (P:program ; m:message)
     (is_RAM_instruction registers_no m P) ->
     ~m=n ->
     (instruction_does_not_occur registers_no n (Program (Action (BangIn m) P)))
 | IOO0Parallel :
    (P,Q:agent)
     (instruction_does_not_occur registers_no n P) ->
      (instruction_does_not_occur registers_no n Q) ->
       (instruction_does_not_occur registers_no n (AParallel P Q)).

Lemma instruction_does_not_occur_decidable:
 (registers_no:nat; n:nat; a:agent)
  {(instruction_does_not_occur registers_no n a)}+
  {~(instruction_does_not_occur registers_no n a)}.
Proof.
 Induction a;
  [ Left;
    Constructor
  | Induction p;
     [ Left;
       Constructor
     | Induction a0;
        [ Right;
          Intro;
          Inversion H0
        | Right;
          Intro;
          Inversion H0
        | Right;
          Intro;
          Inversion H0
        | Intros;
          Elim (eq_nat_dec m n);
           [ Right;
             Intro;
             Inversion H0;
             Apply (H4 a1)
           | Elim (is_RAM_instruction_decidable registers_no m p0);
              [ Left;
                Constructor;
                 [ Assumption
                 | Assumption
                 ]
              | Right;
                Intro;
                Inversion H0;
                Apply (b H3)
              ]
           ]
        ]
     | Right;
       Intro;
       Inversion H1
     | Right;
       Intro;
       Inversion H1
     ]
  | Intros;
    Elim H;
     [ Intros;
       Elim H0;
        [ Intros;
          Left;
          Constructor;
           [ Assumption
           | Assumption
           ]
        | Intros;
          Right;
          Intro;
          Inversion H1;
          Elim (b H5)
        ]
     | Intros;
       Right;
       Intro;
       Inversion H1;
       Apply (b H4)
     ]
  ].
Qed.

(*** Lemmas about instruction_does_not_occur ***)

Lemma instruction_does_not_occur_no_LIn :
 (registers_no:nat ; n:nat ; P,P':agent)
  (instruction_does_not_occur registers_no n P) ->
   ~(OStep P (LIn n) P').
Proof.
 Unfold not;
 Induction P;
  [ Intros;
    Inversion H0
  | Intros;
    Inversion H;
     [ Rewrite <- H2 in H0;
       Inversion H0
     | Rewrite <- H1 in H0;
       Inversion H0;
       Exact (H3 H7)
     ]
  | Intros;
    Inversion H1;
    Inversion H2;
     [ Apply (H ? H5 H12)
     | Apply (H0 ? H6 H12)
     ]
  ].
Qed.


Inductive instruction_occurs_once [registers_no:nat ; n:nat] : agent -> Prop :=
   IOOBangIn :
    (P:program)
     (is_RAM_instruction registers_no n P) ->
     (instruction_occurs_once registers_no n (Program (Action (BangIn n) P)))
 | IOOParallel1 :
    (P,Q:agent)
     (instruction_occurs_once registers_no n P) ->
      (instruction_does_not_occur registers_no n Q) ->
       (instruction_occurs_once registers_no n (AParallel P Q))
 | IOOParallel2 :
    (P,Q:agent)
     (instruction_does_not_occur registers_no n P) ->
      (instruction_occurs_once registers_no n Q) ->
       (instruction_occurs_once registers_no n (AParallel P Q)).

Lemma instruction_does_not_occur_not_instruction_occurs_once:
 (registers_no:nat ; n:nat ; a:agent)
  (instruction_does_not_occur registers_no n a) ->
   ~(instruction_occurs_once registers_no n a).
Proof.
 Intros;
 Elim H;
  [ Intros;
    Intro;
    Inversion H0
  | Intro;
    Inversion H0
  | Intros;
    Intro;
    Inversion H2;
    Apply H1;
    Symmetry;
    Assumption
  | Intros;
    Intro;
    Inversion H4;
     [ Apply (H1 H7)
     | Apply (H3 H8)
     ]
  ].
Qed.

(* The inversion principle in Set *)
Lemma instruction_occurs_once_inv:
 (registers_no,n:nat; a:agent; P:(nat->nat->agent->Set))
  ((instruction_occurs_once registers_no n a)
   ->(P0:program)
      (Program (Action (BangIn n) P0))=a
      ->(is_RAM_instruction registers_no n P0)
      ->(P registers_no n a))
  ->((instruction_occurs_once registers_no n a)
     ->(P0,Q:agent)
        (AParallel P0 Q)=a
        ->(instruction_occurs_once registers_no n P0)
        ->(instruction_does_not_occur registers_no n Q)
        ->(P registers_no n a))
  ->((instruction_occurs_once registers_no n a)
     ->(P0,Q:agent)
        (AParallel P0 Q)=a
        ->(instruction_does_not_occur registers_no n P0)
        ->(instruction_occurs_once registers_no n Q)
        ->(P registers_no n a))
  ->(instruction_occurs_once registers_no n a)
  ->(P registers_no n a).
Proof.
 Intros;
 Induction a;
  [ ElimType False;
    Inversion H2
  | Induction p;
     [ ElimType False;
       Inversion H2
     | Induction a;
        [ ElimType False;
          Inversion H2
        | ElimType False;
          Inversion H2
        | ElimType False;
          Inversion H2
        | Elim (eq_nat_dec n m);
           [ Intro;
             Rewrite <- a;
             Rewrite <- a in H;
             Rewrite <- a in H2;
             Apply (H H2 p (refl_equal ? (Program (Action (BangIn n) p))));
             Inversion H2;
             Assumption
           | Intros;
             Elim b;
             Inversion H2;
             Reflexivity
           ]
        ]
     | ElimType False;
       Inversion H2
     | ElimType False;
       Inversion H2
     ]
  | Elim (instruction_does_not_occur_decidable registers_no n a1);
     [ Intro;
       Apply (H1 H2 a1 a0 (refl_equal ? (AParallel a1 a0)));
        [ Assumption
        | Inversion H2;
           [ Elim
              (instruction_does_not_occur_not_instruction_occurs_once
                ? ? ? a H5)
           | Assumption
           ]
        ]
     | Intro;
       Cut
        (instruction_occurs_once registers_no n a1) /\
         (instruction_does_not_occur registers_no n a0);
        [ Intro;
          Case H3;
          Intros;
          Apply (H0 H2 a1 a0 (refl_equal ? (AParallel a1 a0)));
           [ Assumption
           | Assumption
           ]
        | Inversion H2;
           [ Split;
              [ Assumption
              | Assumption
              ]
           | Elim (b H5)
           ]
        ]
     ]
  ].
Qed.

(*** Lemmas about instruction_occurs_once ***)

Lemma instruction_occurs_once_RAM_image_no_pc_LIn :
 (registers_no:nat ; pc:nat ; P:agent)
  (instruction_occurs_once registers_no pc P) ->
  (is_RAM_image_no_pc registers_no P) ->
   (ge pc registers_no) ->
    {P':agent |
     (OStep P (LIn pc) P') /\
      (EX r |
       (lt r registers_no) /\
        ((is_RAM_image_succ registers_no r pc P') \/
         (EX pc' |
	  (ge pc' registers_no) /\
	   (is_RAM_image_decjump registers_no r pc' pc P'))))}.
Proof.
 Induction P;
  [ Intros;
    ElimType False;
    Inversion H
  | Intros;
    Inversion H using instruction_occurs_once_inv;
     [ Intros;
       Injection H3;
       Intro;
       Rewrite <- H5;
       EApply exist;
       Split;
        [ Apply SBangIn
        | Inversion H4;
           [ Exists r;
             Split;
              [ Assumption
              | Left;
                Constructor;
                 [ Constructor
                 | Constructor;
                    [ Assumption
                    | Constructor;
                      Assumption
                    ]
                 ]
              ]
           | Exists r;
             Split;
              [ Assumption
              | Right;
	        Exists n;
		Split;
		 [ Assumption
		 | Constructor;
	            [ Constructor
	            | Constructor;
	               [ Assumption
		       | Constructor;
		          [ Assumption
                          | Assumption
		          ]
		       ]
	            ]
	         ]
	      ]
           ]
        ]
     | Intros;
       Discriminate H3
     | Intros;
       Discriminate H3
     ]
  | Intros;
    Inversion H1 using instruction_occurs_once_inv;
     [ Intros;
       Inversion H5
     | Intros;
       Inversion H5;
       Rewrite H9 in H6;
       Elim (H H6);
        [ Intros;
          EApply exist;
	  Split;
	   [ Apply SAParallelPL;
              [ Unfold not;
                Induction 1;
                Intros;
                Discriminate
              | Elim p;
	        Intros;
		Apply H8
              ]
	   | Elim p;
	     Intros;
	     Elim H11;
	     Intros;
	     Exists x0;
             Decompose [and] H12;
             Split;
              [ Assumption
              | Elim H14;
	         [ Intro;
	           Left;
		   Constructor;
		    [ Assumption
                    | Inversion H2;
                      Assumption
		    ]
	         | Intro;
                   Right;
                   Elim H15;
                   Intros;
		   Exists x1;
		   Decompose [and] H16;
		   Split;
		    [ Assumption
		    | Constructor;
		       [ Assumption
		       | Inversion H2;
		         Assumption
		       ]
		    ]
	         ]
	      ]
	   ]
        | Inversion H2;
          Assumption
        | Assumption
	]
     | Intros;
       Inversion H5;
       Rewrite H10 in H7;
       Elim (H0 H7);
        [ Intros;
          Elim p;
	  Intros;
	  Exists (AParallel a x);
	  Split;
	   [ Apply SAParallelPR;
              [ Unfold not;
                Induction 1;
                Intros;
                Discriminate
              | Elim p;
	        Intros;
		Assumption
              ]
	   | Elim H11;
	     Intros;
             Decompose [and] H12;
	     Elim H14;
	      [ Intro;
	        Exists x0;
		Split;
                 [ Assumption
                 | Elim H14;
		    [ Intro;
		      Left;
		      Apply IRISAParallel2;
		       [ Inversion H2;
		         Assumption
		       | Assumption
		       ]
		    | Intro;
		      Right;
		      Elim H16;
		      Intros;
		      Exists x1;
		      Decompose [and] H17;
		      Split;
		       [ Assumption
		       | Apply IRIDCAParallel2;
		          [ Inversion H2;
		            Assumption
		          | Assumption
		          ]
                       ]
		    ]
                 ]
	      | Intro;
	        Elim H15;
		Intros;
		Exists x0;
                Split;
                 [ Assumption
                 | Right;
		   Exists x1;
		   Decompose [and] H16;
		   Split;
		    [ Assumption
		    | Apply IRIDCAParallel2;
		       [ Inversion H2;
		         Assumption
		       | Assumption
		       ]
		    ]
                 ]
	      ]
	   ]
        | Inversion H2;
	  Assumption
        | Assumption
	]
     ]
  ].
Qed.

Lemma instruction_occurs_once_LIn_deterministic :
 (registers_no:nat ; n:nat ; P,P',P'':agent)
  (instruction_occurs_once registers_no n P) ->
   (OStep P (LIn n) P') -> (OStep P (LIn n) P'') ->
    P'=P''.
Proof.
 Induction P;
  [ Intros;
    Inversion H0
  | Intros;
    Inversion H;
    Rewrite <- H2 in H0;
    Rewrite <- H2 in H1;
    Inversion H0;
    Inversion H1;
    Reflexivity
  | Intros;
    Inversion H1;
     [ Inversion H2;
        [ Inversion H3;
           [ Elim (H ? ? H6 H13 H19);
             Reflexivity
           | Elim (instruction_does_not_occur_no_LIn ? ? ? ? H7 H19)
           ]
        | Inversion H3;
           [ Elim (instruction_does_not_occur_no_LIn ? ? ? ? H7 H13)
           | Elim (instruction_does_not_occur_no_LIn ? ? ? ? H7 H13)
           ]
        ]
     | Inversion H2;
        [ Inversion H3;
           [ Elim (instruction_does_not_occur_no_LIn ? ? ? ? H6 H13)
           | Elim (instruction_does_not_occur_no_LIn ? ? ? ? H6 H13)
           ]
        | Inversion H3;
           [ Elim (instruction_does_not_occur_no_LIn ? ? ? ? H6 H19)
           | Elim (H0 ? ? H7 H13 H19);
             Reflexivity
           ]
        ]
     ]
  ].
Qed.

Inductive is_RAM_image [registers_no : nat ; program_length : nat ; P:agent] :
 Prop
:=
 Build_is_RAM_image:
  (is_RAM_image_pc registers_no P) ->
   ((n:nat)(lt n program_length) ->
     (instruction_occurs_once registers_no (plus registers_no n) P)) -> 
   ((n:nat)(ge n program_length) ->
     (instruction_does_not_occur registers_no (plus registers_no n) P)) ->
   (is_RAM_image registers_no program_length P).

Lemma is_RAM_image_ind_dependent :
 (registers_no,program_length:nat; P:agent;
  P0:((is_RAM_image registers_no program_length P)->Prop))
   ((i:(is_RAM_image_pc registers_no P);
     i0:
      ((n:nat)
        (lt n program_length)
        ->(instruction_occurs_once registers_no
            (plus registers_no n) P));
     i1:((n:nat)
          (ge n program_length)
          ->(instruction_does_not_occur registers_no
              (plus registers_no n) P)))
     (P0
       (Build_is_RAM_image registers_no program_length P i i0 i1)))
   ->(i:(is_RAM_image registers_no program_length P))(P0 i).
Proof.
 Exact
  [registers_no,program_length:nat;
   P:agent;
   P0:((is_RAM_image registers_no program_length P)->Prop);
   f:
    ((i:(is_RAM_image_pc registers_no P);
      i0:
       ((n:nat)
         (lt n program_length)
         ->(instruction_occurs_once registers_no (plus registers_no n) P));
      i1:((n:nat)
           (ge n program_length)
           ->(instruction_does_not_occur registers_no
               (plus registers_no n) P)))
      (P0 (Build_is_RAM_image registers_no program_length P i i0 i1)));
   i:(is_RAM_image registers_no program_length P)]
   <P0>Cases i of (Build_is_RAM_image x x0 x1) => (f x x0 x1) end.
Qed.

(*** Lemmas about the combination of is_RAM_image[_pc|_no_pc] and ***)
(*** instruction_does_not_occur|instruction_occurs_only_once      ***)

Lemma RAM_image_deterministic_no_LNot :
 (registers_no:nat ; program_length:nat ; P:agent)
  (is_RAM_image registers_no program_length P) ->
    ~(EX m | (EX P' | (OStep P (LNot m) P'))).
Proof.
 Intros;
 Elim H;
 Intros;
 Apply (RAM_image_pc_no_LNot ? ? H0).
Qed.

Theorem is_RAM_image_pc_instruction_occurs_once_Tau :
 (registers_no:nat ; P:agent)
  (H:(is_RAM_image_pc registers_no P))
   [pc := (pc_of_RAM_image_pc ? ? H)]
    (instruction_occurs_once registers_no pc P) ->
     {P':agent |
      (OStep P Tau P') /\
       (EX r |
        (lt r registers_no) /\
         ((is_RAM_image_succ registers_no r pc P') \/
          (EX pc' |
           (ge pc' registers_no) /\
            (is_RAM_image_decjump registers_no r pc' pc P'))))}.
Proof.
 Induction P;
  [ Intros;
    ElimType False;
    Inversion H0
  | Intros;
    ElimType False;
    Inversion H
  | Intros;
    Inversion H1 using is_RAM_image_pc_inv;
     [ Intros;
       Discriminate H4
     | Intros;
       Injection H4;
       Intros;
       Rewrite H8 in H5;
       Inversion H2 using instruction_occurs_once_inv;
        [ Intros;
          Discriminate H10
        | Intros;
          Injection H10;
          Intros;
          Rewrite H14 in H11;
          Elim (H H5) ;
           [ Intros;
             Exists (AParallel x a0);
             Split;
              [ Constructor;
                 [ Unfold not;
                   Induction 1;
                   Intros;
                   Discriminate
                 | Elim p;
                   Intros;
                   Assumption
                 ]
              | Elim p;
                Intros;
                Elim H16;
                Intros;
                Exists x0;
                Decompose [and] H17;
                Split;
                 [ Assumption
                 | Elim H19;
                    [ Intro;
                      Left;
                      Apply IRISAParallel1;
                       [ Unfold pc;
                         Rewrite
                          (pc_of_AParallel_equal_pc1 ? ? ? H1 H5);
                         Assumption
                       | Rewrite <- H7;
                         Assumption
                       ]
                    | Intro;
                      Right;
                      Elim H20;
                      Intros;
                      Exists x1;
                      Decompose [and] H21;
                      Split;
                       [ Assumption
		       | Apply IRIDCAParallel1;
                          [ Unfold pc;
                            Rewrite
                             (pc_of_AParallel_equal_pc1 ? ? ? H1 H5);
                            Assumption
                          | Rewrite <- H7;
                            Assumption
                          ]
                       ]
                    ]
                 ]
              ]
           | Cut pc=(pc_of_RAM_image_pc registers_no a H5);
              [ Intro;
                Rewrite <- H15;
                Assumption
              | Unfold pc;
                Apply pc_of_AParallel_equal_pc1
              ]
           ]
        | Intros;
          Injection H10;
          Intros;
          Rewrite H13 in H12;
          Elim (RAM_image_pc_LOut ? ? H5);
          Intros;
          Rewrite <- (pc_of_AParallel_equal_pc1 ? ? ? H1 H5) in p;
          Fold pc in p;
          Elim (instruction_occurs_once_RAM_image_no_pc_LIn ? ? ? H12);
	   [ Intros;
             Exists (AParallel x x0);
	     Elim p0;
	     Intros;
	     Elim H16;
	     Intros;
             Split;
              [ Apply (SAParallelTauIn2 ? ? ? ? ? p H15)
              | Exists x1;
                Decompose [and] H17;
                Split;
                 [ Assumption
                 | Elim H19;
		    [ Intro;
		      Left;
		      Apply IRISAParallel2;
		       [ Apply
		          RAM_image_pc_LOut_RAM_image_no_pc with pc := pc P := a;
		          [ Apply (pc_of_RAM_image_pc_ge_registers_no ? ? H1)
		          | Assumption
		          | Assumption
		          ]
		       | Assumption
		       ]
		    | Intro;
		      Right;
		      Elim H20;
		      Intros;
                      Exists x2;
		      Decompose [and] H21;
		      Split;
		       [ Assumption
		       | Apply IRIDCAParallel2;
		          [ Apply
		             RAM_image_pc_LOut_RAM_image_no_pc
			      with pc := pc P := a;
		             [ Apply (pc_of_RAM_image_pc_ge_registers_no ? ? H1)
		             | Assumption
		             | Assumption
		             ]
		          | Assumption
		          ]
                       ]
		    ]
                 ]
              ]
	   | Rewrite <- H7;
	     Assumption
	   | Apply (pc_of_RAM_image_pc_ge_registers_no ? ? H1)
	   ]
        ]
     | Inversion H2 using instruction_occurs_once_inv;
        [ Intros;
          Discriminate H4
        | Intros;
          Injection H8;
          Intros;
          Rewrite H12 in H9;
          Rewrite H11 in H10;
          Elim (RAM_image_pc_LOut ? ? H10);
          Intros;
          Rewrite <- (pc_of_AParallel_equal_pc2 ? ? ? H1 H10) in p;
          Fold pc in p;
          Injection H4;
          Intros;
          Rewrite H14 in H5;
          Elim (instruction_occurs_once_RAM_image_no_pc_LIn ? ? ? H5 H9);
	   [ Intros;
             Exists (AParallel x0 x);
	     Elim p0;
	     Intros;
             Split;
              [ Apply (SAParallelTauIn1 ? ? ? ? ? H15 p)
              | Elim H16;
	        Intros;
		Exists x1;
                Decompose [and] H17;
                Split;
                 [ Assumption
                 | Elim H19;
		    [ Intro;
		      Left;
                      Apply IRISAParallel1;
                       [ Assumption
		       | Apply
		          RAM_image_pc_LOut_RAM_image_no_pc with pc := pc P := a0;
		          [ Apply (pc_of_RAM_image_pc_ge_registers_no ? ? H1)
		          | Assumption
		          | Assumption
		          ]
                       ]
		    | Intro;
		      Right;
                      Decompose [ex and] H20;
                      Exists x2;
		      Split;
		       [ Assumption
		       | Apply IRIDCAParallel1;
		          [ Assumption
                          | Apply
		             RAM_image_pc_LOut_RAM_image_no_pc with pc := pc P := a0;
		             [ Apply (pc_of_RAM_image_pc_ge_registers_no ? ? H1)
		             | Assumption
		             | Assumption
		             ]
                          ]
                       ]
		    ]
		 ]
              ]
	   | Apply (pc_of_RAM_image_pc_ge_registers_no ? ? H1)
	   ]
        | Intros;
          Injection H8;
          Intros;
          Injection H4;
          Intros;
          Rewrite H13 in H6;
          Rewrite H11 in H10;
          Elim (H0 H10) ;
           [ Intros;
             Exists (AParallel a x);
             Split;
              [ Constructor;
                 [ Unfold not;
                   Induction 1;
                   Intros;
                   Discriminate
                 | Elim p;
                   Intros;
                   Assumption
                 ]
              | Elim p;
                Intros;
                Elim H16;
                Intros;
                Exists x0;
                Decompose [and] H17;
                Split;
                 [ Assumption
                 | Elim H19;
                    [ Intro;
                      Left;
                      Apply IRISAParallel2;
                       [ Rewrite <- H12;
                         Assumption
                       | Unfold pc;
                         Rewrite
                          (pc_of_AParallel_equal_pc2 ? ? ? H1 H10);
                         Assumption
                       ]
                    | Intro;
                      Right;
                      Elim H20;
                      Intros;
                      Exists x1;
		      Decompose [and] H21;
		      Split;
		       [ Assumption
		       | Apply IRIDCAParallel2;
                          [ Rewrite <- H12;
                            Assumption
                          | Unfold pc;
                            Rewrite
                             (pc_of_AParallel_equal_pc2 ? ? ? H1 H10);
                            Assumption
                          ]
                       ]
                    ]
                 ]
              ]
           |
              Cut pc=(pc_of_RAM_image_pc registers_no a0 H10);
              [ Intro;
                Rewrite <- H15;
                Assumption
              | Unfold pc;
                Apply pc_of_AParallel_equal_pc2
              ]
           ]
        ]
     ]
  ].
Defined.

Theorem is_RAM_image_pc_instruction_does_not_occur_not_Tau :
 (registers_no:nat ; P,P':agent)
  (H:(is_RAM_image_pc registers_no P))
   [pc := (pc_of_RAM_image_pc ? ? H)]
    (instruction_does_not_occur registers_no pc P) ->
     ~(OStep P Tau P').
Proof.
 Unfold not;
 Induction P;
  [ Intros;
    Inversion H1
  | Intros;
    Inversion H
  | Intros;
    Inversion H2;
    Inversion H3;
     [ Inversion H1;
        [ Apply (H P'0 H16);
           [ Rewrite <- (pc_of_AParallel_equal_pc1 ? ? ? H1 H16);
             Assumption
           | Assumption
           ]
        | Apply (stuck_RAM_image_no_pc ? ? H16);
          Exists P'0;
          Left;
          Assumption
        ]
     | Inversion H1;
        [ Apply (stuck_RAM_image_no_pc ? ? H17);
          Exists P'0;
          Left;
          Assumption
        | Apply (H0 P'0 H17);
           [ Rewrite <- (pc_of_AParallel_equal_pc2 ? ? ? H1 H17);
             Assumption
           | Assumption
           ]
        ]
     | Inversion H1;
        [ Generalize
           (RAM_image_pc_LIn_ge_registers_no ? ? ? ? H15 H10);
          Intro Hge;
          Generalize
           (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? H16 H11);
          Intro Hlt;
          Apply (le_not_lt ? ? Hge Hlt)
        | (** This part of the proofs is not at all trivial and **)
          (** requires several lemmas.                          **)
          Generalize
           (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? H15 H10);
          Intro Hge;
          Elim (RAM_image_pc_LOut? ? H16);
          Intros;
          Generalize (pc_of_RAM_image_pc_ge_registers_no ? ? H16);
          Intro;
          Generalize
           (RAM_image_pc_LOut_deterministic ? ? ? Hge H17 ? ? ? H16 H11 p);
          Intro;
          Case H18;
          Clear H18;
          Intros;
          Generalize (pc_of_AParallel_equal_pc2 ? ? ? H1 H16);
          Intro;
          Rewrite H20 in H6;
          Rewrite <- H18 in H6;
          Apply (instruction_does_not_occur_no_LIn ? ? ? ? H6 H10)
        ]
     | Inversion H1;
        [ (** This part of the proofs is not at all trivial and **)
          (** requires several lemmas.                          **)
          Generalize
           (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? H16 H11);
          Intro Hge;
          Elim (RAM_image_pc_LOut? ? H15);
          Intros;
          Generalize (pc_of_RAM_image_pc_ge_registers_no ? ? H15);
          Intro;
          Generalize
           (RAM_image_pc_LOut_deterministic ? ? ? Hge H17 ? ? ? H15 H10 p);
          Intro;
          Case H18;
          Clear H18;
          Intros;
          Generalize (pc_of_AParallel_equal_pc1 ? ? ? H1 H15);
          Intro;
          Rewrite H20 in H7;
          Rewrite <- H18 in H7;
          Apply (instruction_does_not_occur_no_LIn ? ? ? ? H7 H11)
        | Generalize
           (RAM_image_pc_LIn_ge_registers_no ? ? ? ? H16 H11);
          Intro Hge;
          Generalize
           (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? H15 H10);
          Intro Hlt;
          Apply (le_not_lt ? ? Hge Hlt)
        ]
     | Inversion H1; 
        [ Apply (RAM_image_pc_no_LRead ? ? ? ? H15 H10)
        | Apply (RAM_image_no_pc_no_LRead ? ? ? ? H15 H10)
        ]
     | Inversion H1;
        [ Apply (RAM_image_no_pc_no_LRead ? ? ? ? H16 H11)
        | Apply (RAM_image_pc_no_LRead ? ? ? ? H16 H11)
        ]
     ]
  ].
Qed.

Theorem RAM_image_pc_AParallel_LOut_LIn_not_instruction_does_not_occur1:
 (registers_no:nat ; P,P',Q,Q':agent ; m:message)
  (H:(is_RAM_image_pc registers_no (AParallel P Q)))
   (OStep P (LOut m) P') ->
   (OStep Q (LIn m) Q') ->
    [pc := (pc_of_RAM_image_pc ? ? H)]
     ~(instruction_does_not_occur registers_no pc Q).
Proof.
 Intros;
 Intro;
 Inversion H;
  [ Generalize (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? H6 H1);
    Intro Hge;
    Generalize (RAM_image_pc_LOut_pc ? ? ? ? Hge H5 H0);
    Intro;
    Rewrite <- (pc_of_AParallel_equal_pc1 ? ? ? H H5) in H7;
    Fold pc in H7;
    Rewrite H7 in H1;
    Elim (instruction_does_not_occur_no_LIn ? ? ? ? H2 H1)
  | Generalize (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? H5 H0);
    Intro Hlt;
    Generalize (RAM_image_pc_LIn_ge_registers_no ? ? ? ? H6 H1);
    Intro Hge;
    Elim (le_not_lt ? ? Hge Hlt)
  ].
Qed.

Theorem RAM_image_pc_AParallel_LOut_LIn_not_instruction_does_not_occur2:
 (registers_no:nat ; P,P',Q,Q':agent ; m:message)
  (H:(is_RAM_image_pc registers_no (AParallel Q P)))
   (OStep P (LOut m) P') ->
   (OStep Q (LIn m) Q') ->
    [pc := (pc_of_RAM_image_pc ? ? H)]
     ~(instruction_does_not_occur registers_no pc Q).
Proof.
 Intros;
 Intro;
 Inversion H;
  [ Generalize (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? H6 H0);
    Intro Hlt;
    Generalize (RAM_image_pc_LIn_ge_registers_no ? ? ? ? H5 H1);
    Intro Hge;
    Elim (le_not_lt ? ? Hge Hlt)
  | Generalize (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? H5 H1);
    Intro Hge;
    Generalize (RAM_image_pc_LOut_pc ? ? ? ? Hge H6 H0);
    Intro;
    Rewrite <- (pc_of_AParallel_equal_pc2 ? ? ? H H6) in H7;
    Fold pc in H7;
    Rewrite H7 in H1;
    Elim (instruction_does_not_occur_no_LIn ? ? ? ? H2 H1)
  ].
Qed.

Theorem is_RAM_image_pc_instruction_occurs_once_Tau_deterministic :
 (registers_no:nat ; P,P',P'':agent)
  (H:(is_RAM_image_pc registers_no P))
   [pc := (pc_of_RAM_image_pc ? ? H)]
    (instruction_occurs_once registers_no pc P) ->
     (OStep P Tau P') -> (OStep P Tau P'') -> P'=P''.
Proof.
 Induction P;
  [ Intros;
    Inversion H1
  | Intros;
    Inversion H
  | Intros;
    Inversion H1;
     [ Inversion H2;
        [ Inversion H3;
           [ Inversion H4;
              [ Unfold pc in H11;
                Rewrite (pc_of_AParallel_equal_pc1 ? ? ? H1 H7) in H11;
                Rewrite (H ? ? H7 H11 H18 H24);
                Reflexivity
              | Elim (stuck_RAM_image_no_pc ? ? H8);
                Exists P'1;
                Left;
                Assumption
              | Generalize
                 (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? H8 H22) ;
                Intro Hlt;
                Generalize
                 (RAM_image_pc_LIn_ge_registers_no ? ? ? ? H7 H21);
                Intro Hge;
                Elim (le_not_lt ? ? Hge Hlt)
              | Elim
                (RAM_image_pc_AParallel_LOut_LIn_not_instruction_does_not_occur1
                   ? ? ? ? ? ? H1 H21 H22 H12)
              | Elim (RAM_image_pc_no_LRead ? ? ? ? H7 H21)
              | Elim (RAM_image_no_pc_no_LRead ? ? ? ? H8 H22)
              ]
           | Elim (stuck_RAM_image_no_pc ? ? H8);
             Exists P'0;
             Left;
             Assumption
           | Generalize
              (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? H8 H16) ;
             Intro Hlt;
             Generalize
              (RAM_image_pc_LIn_ge_registers_no ? ? ? ? H7 H15);
             Intro Hge;
             Elim (le_not_lt ? ? Hge Hlt)
           | Elim
              (RAM_image_pc_AParallel_LOut_LIn_not_instruction_does_not_occur1
                ? ? ? ? ? ? H1 H15 H16 H12)
           | Elim (RAM_image_pc_no_LRead ? ? ? ? H7 H15)
           | Elim (RAM_image_no_pc_no_LRead ? ? ? ? H8 H16)
           ]
        | Inversion H3;
           [ Unfold pc in H11;
             Rewrite (pc_of_AParallel_equal_pc1 ? ? ? H1 H7) in H11;
             Elim
              (is_RAM_image_pc_instruction_does_not_occur_not_Tau ? ? ?
                H7 H11 H18)
           | Elim (stuck_RAM_image_no_pc ? ? H8);
             Exists P'0;
             Left;
             Assumption
           | Elim
              (RAM_image_pc_AParallel_LOut_LIn_not_instruction_does_not_occur2
                ? ? ? ? ? ? H1 H16 H15 H11)
           | Inversion H4;
              [ Unfold pc in H11;
                Rewrite (pc_of_AParallel_equal_pc1 ? ? ? H1 H7) in H11;
                Elim
                 (is_RAM_image_pc_instruction_does_not_occur_not_Tau ? ? ?
                   H7 H11 H23)
              | Elim (stuck_RAM_image_no_pc ? ? H8);
                Exists P'1;
                Left;
                Assumption
              | Elim
                (RAM_image_pc_AParallel_LOut_LIn_not_instruction_does_not_occur2
                  ? ? ? ? ? ? H1 H21 H20 H11)
              | Generalize
                 (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? H8 H16);
                Intro Hge1;
                Generalize
                 (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? H8 H21);
                Intro Hge2;
                Elim (RAM_image_pc_LOut_deterministic ? ? ? Hge1 Hge2 ? ? ?
                 H7 H15 H20);
                Intros;
                Rewrite <- H24;
                Generalize (RAM_image_pc_LOut_pc ? ? ? ? Hge1 H7 H15);
                Intro Heq;
                Rewrite <- (pc_of_AParallel_equal_pc1 ? ? ? H1 H7) in Heq;
                Fold pc in Heq;
                Rewrite Heq in H16;
                Rewrite <- H23 in H21;
                Rewrite Heq in H21;
                Rewrite
                (instruction_occurs_once_LIn_deterministic ? ? ? ? ?
                  H12 H16 H21);
                Reflexivity
              | Elim (RAM_image_pc_no_LRead ? ? ? ? H7 H20)
              | Elim (RAM_image_no_pc_no_LRead ? ? ? ? H8 H21)
              ]
           | Elim (RAM_image_pc_no_LRead ? ? ? ? H7 H15)
           | Elim (RAM_image_no_pc_no_LRead ? ? ? ? H8 H16)
           ]
        ]
     | Inversion H2;
        [ Inversion H3;
           [ Elim (stuck_RAM_image_no_pc ? ? H7);
             Exists P'0;
             Left;
             Assumption
           | Unfold pc in H12;
             Rewrite (pc_of_AParallel_equal_pc2 ? ? ? H1 H8) in H12;
             Elim
              (is_RAM_image_pc_instruction_does_not_occur_not_Tau ? ? ?
                H8 H12 H18)
           | Inversion H4;
              [ Elim (stuck_RAM_image_no_pc ? ? H7);
                Exists P'1;
                Left;
                Assumption
              | Unfold pc in H12;
                Rewrite (pc_of_AParallel_equal_pc2 ? ? ? H1 H8) in H12;
                Elim
                 (is_RAM_image_pc_instruction_does_not_occur_not_Tau ? ? ?
                   H8 H12 H23)
              | Generalize
                 (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? H7 H20);
                Intro Hge1;
                Generalize
                 (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? H7 H15);
                Intro Hge2;
                Elim (RAM_image_pc_LOut_deterministic ? ? ? Hge1 Hge2 ? ? ?
                 H8 H21 H16);
                Intros;
                Rewrite <- H24;
                Generalize (RAM_image_pc_LOut_pc ? ? ? ? Hge2 H8 H16);
                Intro Heq;
                Rewrite <- (pc_of_AParallel_equal_pc2 ? ? ? H1 H8) in Heq;
                Fold pc in Heq;
                Rewrite Heq in H15;
                Rewrite H23 in H20;
                Rewrite Heq in H20;
                Rewrite
                (instruction_occurs_once_LIn_deterministic ? ? ? ? ?
                  H11 H15 H20);
                Reflexivity
              | Elim
                (RAM_image_pc_AParallel_LOut_LIn_not_instruction_does_not_occur1
                  ? ? ? ? ? ? H1 H20 H21 H12)
              | Elim (RAM_image_no_pc_no_LRead ? ? ? ? H7 H20)
              | Elim (RAM_image_pc_no_LRead ? ? ? ? H8 H21)
              ]
           | Elim
              (RAM_image_pc_AParallel_LOut_LIn_not_instruction_does_not_occur1
                ? ? ? ? ? ? H1 H15 H16 H12)
           | Elim (RAM_image_no_pc_no_LRead ? ? ? ? H7 H15)
           | Elim (RAM_image_pc_no_LRead ? ? ? ? H8 H16)
           ]
        | Inversion H3;
           [ Elim (stuck_RAM_image_no_pc ? ? H7);
             Exists P'0;
             Left;
             Assumption
           | Inversion H4;
              [ Elim (stuck_RAM_image_no_pc ? ? H7);
                Exists P'1;
                Left;
                Assumption
              | Unfold pc in H12;
                Rewrite (pc_of_AParallel_equal_pc2 ? ? ? H1 H8) in H12;
                Rewrite (H0 ? ? H8 H12 H18 H24);
                Reflexivity
              | Elim
                (RAM_image_pc_AParallel_LOut_LIn_not_instruction_does_not_occur2
                   ? ? ? ? ? ? H1 H22 H21 H11)
              | Generalize
                 (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? H7 H21) ;
                Intro Hlt;
                Generalize
                 (RAM_image_pc_LIn_ge_registers_no ? ? ? ? H8 H22);
                Intro Hge;
                Elim (le_not_lt ? ? Hge Hlt)
              | Elim (RAM_image_no_pc_no_LRead ? ? ? ? H7 H21)
              | Elim (RAM_image_pc_no_LRead ? ? ? ? H8 H22)
              ]
           | Elim
              (RAM_image_pc_AParallel_LOut_LIn_not_instruction_does_not_occur2
                ? ? ? ? ? ? H1 H16 H15 H11)
           | Generalize
              (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? H7 H15) ;
             Intro Hlt;
             Generalize
              (RAM_image_pc_LIn_ge_registers_no ? ? ? ? H8 H16);
             Intro Hge;
             Elim (le_not_lt ? ? Hge Hlt)
           | Elim (RAM_image_no_pc_no_LRead ? ? ? ? H7 H15)
           | Elim (RAM_image_pc_no_LRead ? ? ? ? H8 H16)
           ]
        ]
     ]
  ].
Qed.

Definition pc_of_RAM_image :
 (registers_no:nat ; program_length:nat ; P:agent)
  (is_RAM_image registers_no program_length P) -> message.
 Intros;
 Inversion H;
 Exact (pc_of_RAM_image_pc ? ? H0).
Defined.

Theorem pc_of_RAM_image_eq_pc_of_RAM_image_pc :
 (registers_no:nat ; program_length:nat ; P:agent)
  (H1:(is_RAM_image registers_no program_length P))
  (H2:(is_RAM_image_pc registers_no P))
   (pc_of_RAM_image ? ? ? H1)=(pc_of_RAM_image_pc ? ? H2).
Proof.
 Intros registers_no program_length P H1;
 Elim H1 using is_RAM_image_ind_dependent;
 Simpl;
 Intros;
 Apply pc_of_RAM_image_pc_proof_irrelevant.
Qed.

Theorem is_RAM_image_OStep_if_pc_lt_program_length :
 (registers_no:nat ; program_length:nat ; P:agent)
  (H:(is_RAM_image registers_no program_length P))
   [pc := (pc_of_RAM_image ? ? ? H)]
    (lt pc (plus registers_no program_length)) ->
     {P':agent | (OStep P Tau P')}.
Proof.
 Intros registers_no program_length P H;
 Case H;
 Intros;
 Simpl in pc;
 Elim (is_RAM_image_pc_instruction_occurs_once_Tau registers_no P i);
  [ Intros;
    Exists x;
    Elim p;
    Intros;
    Assumption
  | Fold pc;
    Generalize (pc_of_RAM_image_pc_ge_registers_no ? ? i);
    Fold pc;
    Intro Hge;
    Elim (le_witness ? ? Hge);
    Intros;
    Rewrite <- p;
    Apply i0;
    Rewrite <- p in H0;
    Unfold lt;
    Unfold lt in H0;
    CutRewrite (S (plus registers_no x))=(plus registers_no (S x)) in H0;
     [ Apply (simpl_le_plus_l ? ? ? H0)
     | Auto
     ]
  ].
Defined.

Theorem is_RAM_image_not_OStep_if_pc_ge_program_length :
 (registers_no:nat ; program_length:nat ; P,P':agent)
  (H:(is_RAM_image registers_no program_length P))
   [pc := (pc_of_RAM_image ? ? ? H)]
    (ge pc (plus registers_no program_length)) ->
     ~(OStep P Tau P').
Proof.
 Intros registers_no program_length P P' H;
 Elim H using is_RAM_image_ind_dependent;
 Intros;
 Unfold not;
 Intro;
 Simpl in pc;
 Apply (is_RAM_image_pc_instruction_does_not_occur_not_Tau ? ? P' i) ;
  [ Generalize (pc_of_RAM_image_pc_ge_registers_no ? ? i);
    Fold pc;
    Intro Hge;
    Elim (le_witness ? ? Hge);
    Intros;
    Rewrite <- p;
    Apply i1;
    Rewrite <- p in H0;
    Apply (simpl_le_plus_l ? ? ? H0)
  | Assumption
  ].
Qed.

Theorem is_RAM_image_Tau_lt_pc_plus_registers_no_program_length :
 (registers_no:nat ; program_length:nat ; P,P':agent)
  (H:(is_RAM_image registers_no program_length P))
   [pc := (pc_of_RAM_image ? ? ? H)]
    (OStep P Tau P') ->
     (lt pc (plus registers_no program_length)).
Proof.
 Intros;
 Elim (le_lt_dec (plus registers_no program_length) pc);
  [ Intro;
    Elim (is_RAM_image_not_OStep_if_pc_ge_program_length ? ? ? ? H a H0)
  | Intro;
    Assumption
  ].
Qed.

Lemma is_RAM_image_stuck_iff_ge_pc_program_length :
 (registers_no:nat ; program_length:nat ; P:agent)
  (H:(is_RAM_image registers_no program_length P))
   [pc := (pc_of_RAM_image ? ? ? H)]
    (stuck P) <-> (ge pc (plus registers_no program_length)).
Proof.
 Intros;
 Split;
  [ Intro;
    Elim (le_lt_dec (plus registers_no program_length) pc);
     [ Intro;
       Assumption
     | Intro;
       Cut ~(stuck P);
        [ Intro;
          Elim (H1 H0)
        | Unfold stuck TransitionSystems.stuck not;
          Intro;
          Apply H1;
          Generalize (is_RAM_image_OStep_if_pc_lt_program_length ? ? ? H b);
          Intro;
          Elim H2;
          Intros;
          Exists x;
          Left;
          Assumption
        ]
     ]
  | Intro;
    Intro;
    Elim H1;
    Intros;
    Apply (is_RAM_image_not_OStep_if_pc_ge_program_length ? ? ? x H H0);
    Elim H2;
     [ Intro;
       Assumption
     | Intro;
       Elim H3;
       Intros;
       Inversion H;
       Elim (RAM_image_pc_no_LNot ? ? H5);
       Exists x0;
       Exists x;
       Assumption
     ]
  ].
Qed.

(*** For RAM images step is deterministic and stuck is decidable ***)

Theorem is_RAM_image_stuck_decidable:
 (registers_no:nat ; program_length:nat ; P:agent)
  (is_RAM_image registers_no program_length P) ->
   {(stuck P)}+{~(stuck P)}.
Proof.
 Intros;
 Elim (is_RAM_image_stuck_iff_ge_pc_program_length ? ? ? H);
 Intros;
 LetTac pc := (pc_of_RAM_image ? ? ? H);
 Elim (le_lt_dec (plus registers_no program_length) pc);
  [ Intro;
    Left;
    Exact (H1 a)
  | Intro;
    Right;
    Intro;
    Generalize (H0 H2);
    Intro;
    Apply (le_not_lt ? ? H3 b)
  ].
Defined.

Theorem RAM_image_deterministic :
 (registers_no:nat ; program_length:nat ; P,P',P'':agent)
  (is_RAM_image registers_no program_length P) ->
   (step P P') -> (step P P'') -> P'=P''.
Proof.
 Intros;
 Elim H using is_RAM_image_ind_dependent;
 Intros;
 Elim H0;
  [ Clear H0;
    Intros;
    Elim H1;
     [ Clear H1;
       Intros;
       Apply
        (is_RAM_image_pc_instruction_occurs_once_Tau_deterministic
          ? ? P' P'' i);
        [ Generalize (pc_of_RAM_image_pc_ge_registers_no ? ? i);
          Intro Hge;
          Elim (le_witness ? ? Hge);
          Intros;
          Rewrite <- p;
          Apply i0;
          Generalize
           (is_RAM_image_Tau_lt_pc_plus_registers_no_program_length ? ? ? ?
             H H0);
          Intro;
          Rewrite (pc_of_RAM_image_eq_pc_of_RAM_image_pc ? ? ? H i) in H2;
          Rewrite <- p in H2;
          Apply (simpl_lt_plus_l ? ? ? H2)
        | Assumption
        | Assumption
        ]
     | Intro;
       Elim (RAM_image_deterministic_no_LNot ? ? ? H);
       Elim H2;
       Intros;
       Exists x;
       Exists P'';
       Assumption
     ]
  | Intro;
    Elim (RAM_image_deterministic_no_LNot ? ? ? H);
    Elim H2;
    Intros;
    Exists x;
    Exists P';
    Assumption
  ].
Qed.

(*** The functions to map back a Linda RAM to a RAM ***)

Definition message_multiplicity : nat -> agent -> nat.
 Intros register P;
 Induction P;
  [ Intros;
    Elim (eq_nat_dec register m);
     [ Intro;
       Exact (S O)
     | Intro;
       Exact O
     ]
  | Intros;
    Exact O
  | Intros;
    Exact (plus HrecP1 HrecP0)
  ].
Defined.

Definition registers_of_agent:
 nat -> agent -> (nat -> nat).
 Intros registers_no P r;
 Elim (le_lt_dec registers_no r);
  [ Intro;
    Exact O  (* dummy value *)
  | Intro;
    Exact (message_multiplicity r P)
  ].
Defined.

Definition configuration_of_RAM_image_pc :
 (registers_no:nat ; P:agent)
  (is_RAM_image_pc registers_no P) -> RAMs.configuration.
 Intros;
 Constructor;
  [ Exact (minus (pc_of_RAM_image_pc ? ? H) registers_no)
  | Exact (registers_of_agent registers_no P)
  ].
Defined.

(*** Basic lemmas about the functions to map back a Linda RAM to a RAM ***)

Lemma message_multiplicity_n_n: 
 (n:message)
  (message_multiplicity n (Message n)) = (1).
Proof.
 Simpl;
 Intro;
 Elim (eq_nat_dec n n);
  [ Reflexivity
  | Intro;
    Elim b;
    Reflexivity
  ].
Qed.

Lemma message_multiplicty_AParallel_O_message_multiplicity_O_L:
 (r:message ; P,Q:agent)
  (message_multiplicity r (AParallel P Q))=(0) ->
   (message_multiplicity r P)=(0).
Proof.
 Intros;
 Simpl in H;
 Exact (MiscRsa.plus_eqO ? ? H).
Qed.

Lemma message_multiplicty_AParallel_O_message_multiplicity_O_R:
 (r:message ; P,Q:agent)
  (message_multiplicity r (AParallel P Q))=(0) ->
   (message_multiplicity r Q)=(0).
Proof.
 Intros;
 Simpl in H;
 Rewrite plus_sym in H;
 Exact (MiscRsa.plus_eqO ? ? H).
Qed.

Lemma message_multiplicity_O_NoMessage:
 (r:message ; Q:agent)
  (message_multiplicity r Q)=(0) -> (NoMessage r Q).
Proof.
 Induction Q;
  [ Simpl;
    Intro;
    Case (eq_nat_dec r m);
     [ Simpl;
       Intros;
       Inversion H
     | Simpl;
       Constructor;
       Assumption
     ]
  | Constructor
  | Constructor;
     [ Apply H;
       Exact
        (message_multiplicty_AParallel_O_message_multiplicity_O_L r a a0 H1)
     | Apply H0;
       Exact
        (message_multiplicty_AParallel_O_message_multiplicity_O_R r a a0 H1)
     ]
  ].
Qed.

Lemma NoMessage_message_multiplicity_O:
 (r:message ; Q:agent)
  (NoMessage r Q) -> (message_multiplicity r Q)=(0).
Proof.
 Induction 1;
  [ Intros;
    Simpl;
    Elim (eq_nat_dec r m');
     [ Intros;
       Elim (H0 a)
     | Reflexivity
     ]
  | Reflexivity
  | Intros;
    Simpl;
    Rewrite H1;
    Rewrite H3;
    Reflexivity
  ].
Qed.

Lemma LOut_not_message_multiplicity_eq_O:
 (m:message ; P,P': agent)
  (OStep P (LOut m) P') ->
   ~(message_multiplicity m P)=(0).
Proof.
 Induction P;
  [ Intros;
    Intro;
    Inversion H;
    Rewrite H3 in H0;
    Rewrite (message_multiplicity_n_n m) in H0;
    Discriminate
  | Intros;
    Inversion H
  | Intros;
    Inversion H1;
     [ Intro;
       Apply (H ? H7);
       Exact (message_multiplicty_AParallel_O_message_multiplicity_O_L ? ? ? H8)
     | Intro;
       Apply (H0 ? H7);
       Exact (message_multiplicty_AParallel_O_message_multiplicity_O_R ? ? ? H8)
     ]
  ].
Qed.

Lemma LOut_message_multiplicity_decr:
 (m:message ; P,P': agent)
  (OStep P (LOut m) P') ->
   (message_multiplicity m P)=(S (message_multiplicity m P')).
Proof.
 Induction P;
  [ Intros;
    Inversion H;
    Rewrite (message_multiplicity_n_n m);
    Reflexivity
  | Intros;
    Inversion H
  | Intros;
    Simpl;
    Inversion H1;
     [ Rewrite (H ? H7);
       Reflexivity
     | Rewrite (H0 ? H7);
       Simpl;
       Rewrite plus_n_Sm;
       Reflexivity
     ]
  ].
Qed.

Lemma not_message_multiplicty_AParallel_O_not_message_multiplicity:
 (r:message ; P,Q:agent)
  ~(message_multiplicity r (AParallel P Q))=(0) ->
   {~(message_multiplicity r P)=(0)}+{~(message_multiplicity r Q)=(0)}.
Proof.
 Intros;
 Simpl in H;
 Apply plus_neq_O_neq_or_neq;
 Assumption.
Qed.

Lemma message_multiplicity_neq_O_RAM_image_no_pc_LOut:
 (registers_no:nat ; r:message ; Q:agent)
  ~(message_multiplicity r Q)=(0) ->
   (is_RAM_image_no_pc registers_no Q) ->
    {Q':agent |
      (is_RAM_image_no_pc registers_no Q') /\
      ((OStep Q (LOut r) Q') /\
        (plus (message_multiplicity r Q') (1)) = (message_multiplicity r Q))}.
Proof.
 Intros;
 Induction Q;
  [ Exists (Program Dead);
    Cut r=m;
     [ Intro;
       Rewrite H1;
       Split;
        [ Constructor
        | Split;
           [ Constructor
           | Rewrite (message_multiplicity_n_n m);
             Reflexivity
           ]
        ]
     | Generalize H;
       Simpl;
       Elim (eq_nat_dec r m);
        [ Trivial
        | Simpl;
          Intros;
          Elim H1;
          Reflexivity
        ]
     ]
  | Elim H;
    Reflexivity
  | Generalize
     (not_message_multiplicty_AParallel_O_not_message_multiplicity r Q1 Q0 H);
    Intro;
    Case H1;
     [ Intro;
       Elim (HrecQ1 n);
        [ Intros;
          Case p;
          Intros;
          Exists (AParallel x Q0);
          Case H3;
          Intros;
          Split;
           [ Apply IRI0AParallel1;
              [ Assumption
              | Inversion H0;
                Assumption
              ]
           | Split;
              [ Apply SAParallelPL;
                 [ Intro;
                   Case H6;
                   Intros;
                   Discriminate
                 | Assumption
                 ]
              | Simpl;
                Rewrite <- H5;
                Rewrite <- plus_assoc_l with p := (1);
                Rewrite <- plus_assoc_l with m := (1);
                Apply (f_equal ? ? [y](plus (message_multiplicity r x) y));
                Apply plus_sym
              ]
           ]
        | Inversion H0;
          Assumption
        ]
     | Intro;
       Elim (HrecQ0 n);
        [ Intros;
          Case p;
          Intros;
          Case H3;
          Intros;
          Exists (AParallel Q1 x);
          Split;
           [ Constructor;
              [ Inversion H0;
                Assumption
              | Assumption
              ]
           | Split;
              [ Apply SAParallelPR;
                 [ Intro;
                   Case H6;
                   Intros;
                   Discriminate
                 | Assumption
                 ]
              | Simpl;
                Rewrite <- H5;
                Rewrite <- plus_assoc_l with p := (1);
                Reflexivity
              ]
           ]
        | Inversion H0;
          Assumption
        ]
     ]
  ].
Qed.

(*** The codomain of the mapping from RAMs to Linda is the set of RAM_images **)

Lemma agent_of_configuration_is_RAM_image_pc :
 (r:RAM ; c:configuration)
  [registers_no := (registers_number r)]
  [program_length := (length (RAMs.program r))]
   (is_RAM_image_pc registers_no
     (agent_of_configuration program_length registers_no c)).
Proof.
 Intros;
 Unfold agent_of_configuration;
 Elim c;
 Clear c;
 Intros;
 Apply IRIAParallel1;
  [ Constructor;
    Unfold ge;
    Apply le_plus_l
  | Apply agent_of_registers_ind;
     [ Constructor
     | Intros;
       Constructor;
        [ Elim (registers r0);
           [ Constructor
           | Intros;
             Simpl;
             Constructor;
              [ Constructor;
                Assumption
              | Assumption
              ]
           ]
        | Assumption
        ]
     ]
  ].
Defined.

(*** agent_of_configuration and configuration_of_RAM_image_pc are ***)
(*** extensionally inverse functions                              ***)

Lemma message_multiplicity_agent_of_register1:
 (r:register ; v:nat) 
  (message_multiplicity r (agent_of_register r v)) = v.
Proof.
 Induction v;
  [ Reflexivity
  | Intros;
    Simpl;
    Elim (eq_nat_dec r r);
     [ Intro;
       Simpl;
       Rewrite H;
       Reflexivity
     | Intro;
       Elim (b (refl_equal ? r))
     ]
  ].
Qed.

Lemma message_multiplicity_agent_of_register2:
 (r,r':register ; v:nat) 
   ~r'=r ->
    (message_multiplicity r' (agent_of_register r v)) = O.
Proof.
 Intros;
 Elim v;
  [ Reflexivity
  | Intros;
    Simpl;
    Elim (eq_nat_dec r' r);
     [ Intro;
       Elim (H a)
     | Intro;
       Simpl;
       Assumption
     ]
  ].
Qed.

Lemma message_multiplicity_n_agent_of_registers_n_O :
 (registers: nat -> nat ; n,m:nat)
  (ge m n) ->
   (message_multiplicity m (agent_of_registers registers n)) = O.
Proof.
 Induction n;
  [ Reflexivity
  | Intros;
    Simpl;
    Cut ~m=n0;
     [ Intro;
       Rewrite (message_multiplicity_agent_of_register2 ? ? (registers n0) H1);
       Simpl;
       Apply H;
       Unfold ge;
       Apply Le.le_trans_S;
       Exact H0
     | Intro;
       Rewrite H1 in H0;
       Unfold ge in H0;
       Exact (le_Sn_n ? H0)
     ]
  ].
Qed.

Theorem configuration_of_RAM_image_pc_inverse_agent_of_configuration :
 (r:RAM ; c:configuration)
  [registers_no := (registers_number r)]
  [program_length := (length (RAMs.program r))]
   (RAM_configuration_equality registers_no
    (configuration_of_RAM_image_pc registers_no
     (agent_of_configuration program_length registers_no c)
     (agent_of_configuration_is_RAM_image_pc r c)
    ) c).
Proof.
 Intros;
 Elim c;
 Clear c;
 Simpl;
 Unfold configuration_of_RAM_image_pc;
 Intros;
 Split;
  [ Simpl;
    Elim (le_lt_dec registers_no (plus registers_no program_counter));
     [ Simpl;
       Intro;
       Apply minus_plus
     | Intro;
       Cut (le registers_no (plus registers_no program_counter));
        [ Intro;
          Elim
           (le_not_lt registers_no (plus registers_no program_counter) H b)
        | Apply le_plus_l
        ]
     ]
  | Unfold registers_of_agent;
    Simpl;
    Intro;
    Elim (le_lt_dec registers_no r0);
     [ Intro;
       Simpl;
       Intro;
       Elim (le_not_lt ? ? a H)
     | Intros;
       Simpl;
       Elim (eq_nat_dec r0 (plus registers_no program_counter));
        [ Intro;
          Rewrite a in b;
          Elim (not_lt_plus ? ? b)
        | Intro;
          Simpl;
          Generalize b;
          Elim registers_no;
           [ Intro;
             Simpl;
             Inversion b1
           | Intro;
             Elim (eq_nat_dec r0 n);
              [ Intros;
                Rewrite a;
                Simpl;
                Rewrite
                 (message_multiplicity_agent_of_register1 n (registers n));
                Rewrite 
                 (message_multiplicity_n_agent_of_registers_n_O registers
                   ? ? (le_n n));
                Symmetry;
                Apply plus_n_O
              | Intros;
                Simpl;
                Rewrite
                 (message_multiplicity_agent_of_register2 ? ? (registers n) b1);
                Simpl;
                Apply H0;
                Unfold lt in b2;
                Inversion b2;
                 [ Elim (b1 H2)
                 | Exact H2
                 ]
              ]
           ]
        ]
     ]
  ].
Qed.

Lemma is_RAM_image_no_pc_agent_of_registers:
 (registers: register -> nat ; registers_number: nat ; n: nat)
   (le n registers_number) ->
    (is_RAM_image_no_pc registers_number (agent_of_registers registers n)).
Proof.
 Induction n;
  [ Constructor
  | Intros;
    Simpl;
    Constructor;
     [ Elim (registers n0);
        [ Constructor
        | Intros;
          Simpl;
          Constructor;
           [ Constructor;
             Assumption
           | Assumption
           ]
        ]
     | Apply H;
       Exact (le_trans_S ? ? H0)
     ]
  ].
Qed.

Lemma is_RAM_image_no_pc_agent_of_program:
 (registers_no: nat ; program : (list instruction))
  (AllS (uses_at_most registers_no) program) ->
   (is_RAM_image_no_pc registers_no (agent_of_program registers_no program)).
Proof.
 Intros registers_no program;
 Unfold agent_of_program;
 Cut
  (n:nat ; P:agent)
   (is_RAM_image_no_pc registers_no P) ->
    (AllS (uses_at_most registers_no) program) ->
     (is_RAM_image_no_pc registers_no
       (fold_left_i agent instruction
         ([pc:nat; a:agent; i:instruction]
           (AParallel a
             (Program (program_of_instruction registers_no pc i))))
         program n P));
  [ Intros;
    Apply H;
     [ Constructor
     | Assumption
     ] 
  | Elim program;
     [ Simpl;
       Intros;
       Assumption
     | Simpl;
       Intros;
       Apply H;
        [ Constructor;
           [ Assumption
           | Induction a;
              [ Simpl;
                Constructor;
                 [ Exact (le_plus_l registers_no n)
                 | Rewrite <- (plus_n_Sm registers_no n);
                   Constructor;
                   Inversion H1;
                   Inversion H4;
                   Assumption
                 ]
              | Simpl;
                Constructor;
                 [ Exact (le_plus_l registers_no n)
                 | Rewrite <- (plus_n_Sm registers_no n);
                   Constructor;
                    [ Inversion H1;
                      Inversion H4;
                      Assumption
                    | Exact (le_plus_l registers_no n0)
                    ]
                 ]
              ]
           ]
        | Inversion H1;
          Assumption
        ]
     ]
  ].
Qed.

Lemma instruction_does_not_occur_agent_of_registers :
 (registers_number:nat ; n,m:nat ; registers: register -> nat)
  (instruction_does_not_occur n m
    (agent_of_registers registers registers_number)).
Proof.
 Intros;
 Elim registers_number;
  [ Constructor
  | Intros;
    Simpl;
    Constructor;
     [ Elim (registers n0);
        [ Constructor
        | Simpl;
          Constructor;
           [ Constructor
           | Assumption
           ]
        ]
     | Assumption
     ]
  ].
Qed.

Lemma instruction_does_not_occur_program_of_instruction1 :
 (registers_no:nat ; n:nat ; program: (list instruction); m:nat ; P:agent)
  (ge n (plus (length program) m)) ->
   (AllS (uses_at_most registers_no) program) ->
   (instruction_does_not_occur registers_no (plus registers_no n) P) ->
    (instruction_does_not_occur registers_no (plus registers_no n)
     (fold_left_i agent instruction
       ([pc:nat; a:agent; i:instruction]
         (AParallel a
           (Program (program_of_instruction registers_no pc i))))
       program m P)).
Proof.
 Induction program;
  [ Intros;
    Assumption
  | Intros;
    Simpl;
    Apply H;
     [ Rewrite <- plus_n_Sm;
       Assumption
     | Inversion H1;
       Assumption
     | Constructor;
        [ Assumption
        | Induction a;
           [ Simpl;
             Rewrite <- plus_n_Sm;
             Constructor;
              [ Constructor;
                Inversion H1;
                Inversion H5;
                Assumption
              | Intro;
                Rewrite (simpl_plus_l ? ? ? H3) in H0;
                Simpl in H0;
                Cut (lt n n);
                 [ Apply lt_n_n
                 | Apply (lt_le_trans n (S (plus (length l) n)) n);
                    [ Unfold lt;
                      Apply le_n_S;
                      Apply le_plus_r
                    | Assumption
                    ]
                 ]
              ]
           | Simpl;
             Constructor;
              [ Rewrite <- plus_n_Sm;
                Constructor;
                 [ Inversion H1;
                   Inversion H5;
                   Assumption
                 | Unfold ge;
                   Apply le_plus_l
                 ]
              | Intro;
                Rewrite (simpl_plus_l ? ? ? H3) in H0;
                Simpl in H0;
                Cut (lt n n);
                 [ Apply lt_n_n
                 | Apply (lt_le_trans n (S (plus (length l) n)) n);
                    [ Unfold lt;
                      Apply le_n_S;
                      Apply le_plus_r
                    | Assumption
                    ]
                 ]
              ]
           ]
        ]
     ]
  ].
Qed.

Lemma instruction_does_not_occur_ge_program_length_agent_of_program:
 (registers_no:nat ; n:nat ; program: (list instruction))
  (AllS (uses_at_most registers_no) program) ->
   (ge n (length program)) ->
    (instruction_does_not_occur registers_no
      (plus registers_no n)
      (agent_of_program registers_no program)).
Proof.
 Intros;
 Unfold agent_of_program;
 Apply instruction_does_not_occur_program_of_instruction1;
  [ Rewrite plus_sym;
    Assumption
  | Assumption
  | Constructor
  ].
Qed.

Lemma instruction_does_not_occur_not_eq_program_of_instruction:
 (registers_no:nat ; n,m:nat ; a:instruction)
  ~n=m ->
   (uses_at_most registers_no a) ->
    (instruction_does_not_occur registers_no (plus registers_no n)
      (Program (program_of_instruction registers_no m a))).
Proof.
 Induction a; 
  [ Simpl;
    Constructor;
     [ Rewrite <- plus_n_Sm;
       Constructor;
       Inversion H0;
       Assumption
     | Intro;
       Apply H;
       Symmetry;
       Exact (simpl_plus_l ? ? ? H1)
     ]
  | Simpl;
    Constructor;
     [ Rewrite <- plus_n_Sm;
       Constructor;
        [ Inversion H0;
          Assumption
        | Unfold ge;
          Apply le_plus_l
        ]
     | Intro;
       Apply H;
       Symmetry;
       Exact (simpl_plus_l ? ? ? H1)
     ]
  ].
Qed.

Lemma instruction_occurs_once_eq_n_program_length_agent_of_program :
 (registers_no:nat ; n:nat ; program:(list instruction))
  (a:instruction ; P:agent)
   (AllS (uses_at_most registers_no) program) ->
    (instruction_does_not_occur registers_no (plus registers_no n) P) ->
    (uses_at_most registers_no a) ->
     (instruction_occurs_once registers_no (plus registers_no n)
       (fold_left_i agent instruction
         ([pc:nat; a0:agent; i:instruction]
           (AParallel a0
             (Program (program_of_instruction registers_no pc i)))) program
         (S n)
         (AParallel P (Program (program_of_instruction registers_no n a))))).
Proof.
 Intros;
 Cut
  (m:nat ; P:agent)
   (gt m n) ->
    (instruction_occurs_once registers_no (plus registers_no n) P) ->
     (instruction_occurs_once registers_no (plus registers_no n)
      (fold_left_i agent instruction
        ([pc:nat; a0:agent; i:instruction]
          (AParallel a0
            (Program (program_of_instruction registers_no pc i))))
        program m P));
  [ Intros;
    Apply H2;
     [ Exact (le_n (S n))
     | Apply IOOParallel2;
        [ Assumption
        | Induction a;
           [ Simpl;
             Rewrite <- plus_n_Sm;
             Constructor;
             Constructor;
             Inversion H1;
             Assumption
           | Simpl;
             Rewrite <- plus_n_Sm;
             Constructor;
             Constructor;
              [ Inversion H1;
                Assumption
              | Unfold ge;
                Apply le_plus_l
              ]
           ]
        ]
     ]
  | Induction program;
     [ Intros;
       Assumption
     | Intros;
       Simpl;
       Apply Hrecprogram;
        [ Inversion H;
          Assumption
        | Apply (le_S ? ? H2)
        | Apply IOOParallel1;
           [ Assumption
           | Apply instruction_does_not_occur_not_eq_program_of_instruction;
              [ Intro;
                Rewrite H4 in H2;
                Exact (lt_n_n ? H2)
              | Inversion H;
                Assumption
              ]
           ]
        ]
     ]
  ].
Qed.

Lemma instruction_occurs_once_lt_n_program_length_agent_of_program :
 (registers_no: nat ; n:nat ; program: (list instruction))
  (AllS (uses_at_most registers_no) program) ->
   (lt n (length program)) ->
    (instruction_occurs_once registers_no (plus registers_no n)
     (agent_of_program registers_no program)).
Proof.
 Intros registers_no n program H;
 Unfold agent_of_program;
 Cut
  (m:nat ; P:agent)
   (plus m (length program))=(length program) ->
   (lt n (plus m (length program))) ->
   (ge n m) ->
   (AllS (uses_at_most registers_no) program) ->
   (instruction_does_not_occur registers_no (plus registers_no n) P) ->
    (instruction_occurs_once registers_no (plus registers_no n)
      (fold_left_i agent instruction
        ([pc:nat; a:agent; i:instruction]
          (AParallel a
            (Program (program_of_instruction registers_no pc i))))
        program m P));
  [ Intros;
    Apply H0;
     [ Reflexivity
     | Exact H1
     | Exact (le_O_n n)
     | Assumption
     | Constructor
     ]
  | LetTac k:=(length (program));
    Unfold 1 3 k;
    Elim program;
     [ Intros;
       Simpl in H1;
       Simpl in H0;
       Rewrite -> H0 in H1;
       Unfold k in H1;
       Rewrite -> plus_sym in H0;
       Simpl in H0;
       Rewrite -> H0 in H2;
       Fold k in H1;
       ElimType False;
       Apply (le_not_lt (?) (?) (H2) (H1))
     | Intros;
       Cut {n=m}+{(gt n m)};
        [ Intro;
          Case H6;
           [ Intro;
             Rewrite <- e;
             Simpl;
             Apply instruction_occurs_once_eq_n_program_length_agent_of_program;
              [ Inversion H4;
                Assumption
              | Assumption
              | Inversion H4;
                Assumption
              ]
           | Intros;
             Simpl;
             Apply H0;
              [ Simpl in H1;
                Rewrite <- plus_n_Sm in H1;
                Assumption
              | Simpl in H2;
                Rewrite <- plus_n_Sm in H2;
                Assumption
              | Exact g
              | Inversion H4;
                Assumption
              | Constructor;
                 [ Assumption
                 | Induction a;
                    [ Simpl;
                      Constructor;
                       [ Rewrite <- plus_n_Sm;
                         Constructor;
                         Inversion H4;
                         Inversion H9;
                         Assumption
                       | Intro;
                         Generalize (simpl_plus_l ? ? ? H7);
                         Intro;
                         Apply (gt_neq ? ? g);
                         Symmetry;
                         Assumption
                       ]
                    | Simpl;
                      Constructor;
                       [ Rewrite <- plus_n_Sm;
                         Constructor;
                          [ Inversion H4;
                            Inversion H9;
                            Assumption
                          | Unfold ge;
                            Apply le_plus_l
                          ]
                       | Intro;
                         Generalize (simpl_plus_l ? ? ? H7);
                         Intro;
                         Apply (gt_neq ? ? g);
                         Symmetry;
                         Assumption
                       ]
                    ] 
                 ]
              ]
           ]
        | Exact (ge_eq_sumbool_gt ? ? H3)
        ]
     ]
  ].
Qed.

Lemma agent_of_RAM_and_configuration_is_RAM_image :
 (r:RAM ; c:configuration)
  [registers_no := (registers_number r)]
  [program_length := (length (RAMs.program r))]
   (is_RAM_image registers_no program_length
     (agent_of_RAM_and_configuration r c)).
Proof.
 Intro;
 Case r;
 Clear r;
 Simpl;
 Intros;
 Case c;
 Clear c;
 Simpl;
 Constructor;
  [ Apply IRIAParallel2 ;
     [ Apply is_RAM_image_no_pc_agent_of_program;
       Assumption
     | Apply IRIAParallel1;
        [ Constructor;
          Unfold ge;
          Apply le_plus_l
        | Apply is_RAM_image_no_pc_agent_of_registers;
          Constructor
        ]
     ]
  | Constructor;
     [ Apply instruction_occurs_once_lt_n_program_length_agent_of_program;
       Assumption
     | Constructor;
        [ Constructor
        | Apply instruction_does_not_occur_agent_of_registers
        ]
     ]
  | Constructor;
     [ Apply instruction_does_not_occur_ge_program_length_agent_of_program;
        [ Assumption
        | Assumption
        ]
     | Constructor;
        [ Constructor
        | Apply instruction_does_not_occur_agent_of_registers
        ]
     ]
  ].
Qed.

Lemma pc_of_RAM_image_agent_of_RAM_and_configuration_eq_plus_registerno_program_length:
 (r: RAM ; c: configuration)
  (H:(is_RAM_image (registers_number r) (length (RAMs.program r))
       (agent_of_RAM_and_configuration r c)))
   (pc_of_RAM_image (registers_number r) (length (RAMs.program r))
     (agent_of_RAM_and_configuration r c) H) =
   (plus (registers_number r) (program_counter c)).
Proof.
 Induction c;
 Case r;
 Simpl;
 Intros;
 Elim H using is_RAM_image_ind_dependent;
 Intros;
 Unfold pc_of_RAM_image;
 Cut
  (is_RAM_image_pc registers_number
    (AParallel
      (Message (plus registers_number program_counter))
      (agent_of_registers registers registers_number)));
  [ Intro;
    Rewrite
     (pc_of_AParallel_equal_pc2 registers_number
       (agent_of_program registers_number program)
       (AParallel (Message (plus registers_number program_counter))
         (agent_of_registers registers registers_number)) i H0);
    Cut
     (is_RAM_image_pc registers_number
       (Message (plus registers_number program_counter)));
     [ Intro;
       Rewrite
        (pc_of_AParallel_equal_pc1 registers_number
          (Message (plus registers_number program_counter))
          (agent_of_registers registers registers_number) H0 H1);
       Reflexivity
     | Constructor;
       Unfold ge;
       Apply le_plus_l
     ]
  | Constructor;
     [ Constructor;
       Unfold ge;
       Apply le_plus_l
     | Apply is_RAM_image_no_pc_agent_of_registers;
       Constructor
     ]
  ].
Qed.

(*** Key lemma that ends the first move: a RAM is stuck iff ***)
(*** its Linda image is.                                    ***)

Theorem RAM_stuck_iff_Linda_stuck :
 (r:RAM ; c:configuration)
  (RAMs.stuck r c) <->
  (Linda.stuck (agent_of_RAM_and_configuration r c)).
Proof.
 Intros;
 Generalize (stuck_iff_ge_pc_program_length r c);
 Intro;
 Apply (iff_trans ? ? (stuck (agent_of_RAM_and_configuration r c)) H);
 Generalize (agent_of_RAM_and_configuration_is_RAM_image r c);
 Intro;
 Simpl in H0;
 Generalize (is_RAM_image_stuck_iff_ge_pc_program_length ? ? ? H0);
 Intros;
 Apply iff_sym;
 Apply (iff_trans ? ? (ge (program_counter c) (length (RAMs.program r))) H1);
 Rewrite pc_of_RAM_image_agent_of_RAM_and_configuration_eq_plus_registerno_program_length;
 Split;
  [ Unfold ge;
    Apply simpl_le_plus_l
  | Intro;
    Unfold ge;
    Apply le_reg_l;
    Assumption
  ].
Qed.

(*** Lemmas about is_RAM_image_succ ***)

Lemma RAM_image_succ_not_LNot:
 (registers_no:nat ; r:register ; pc:nat ; m:message ; P,P':agent)
  (is_RAM_image_succ registers_no r pc P) ->
   ~(OStep P (LNot m) P').
Proof.
 Unfold not;
 Intros;
 Generalize m P' H0;
 Elim H;
  [ Intros;
    Inversion H1
  | Intros;
    Inversion H4;
     [ Apply H7;
       Exists m0;
       Reflexivity
     | Apply (RAM_image_no_pc_no_LNot ? ? ? ? H3 H10)
     | Exact (H2 ? ? H8)
     | Apply (RAM_image_no_pc_no_LNot ? ? ? ? H3 H8)
     ]
  | Intros;
    Inversion H4;
     [ Apply (RAM_image_no_pc_no_LNot ? ? ? ? H1 H10)
     | Apply H7;
       Exists m0;
       Reflexivity
     | Apply (RAM_image_no_pc_no_LNot ? ? ? ? H1 H8)
     | Exact (H3 ? ? H8)
     ]
  ].
Qed.

Lemma RAM_image_succ_not_LRead:
 (registers_no:nat ; r:register ; pc:nat ; m:message ; P,P':agent)
  (is_RAM_image_succ registers_no r pc P) ->
   ~(OStep P (LRead m) P').
Proof.
 Unfold not;
 Intros;
 Generalize m P' H0;
 Elim H;
  [ Intros;
    Inversion H1
  | Intros;
    Inversion H4;
     [ Exact (H2 ? ? H10)
     | Apply (RAM_image_no_pc_no_LRead ? ? ? ? H3 H10)
     ]
  | Intros;
    Inversion H4;
     [ Apply (RAM_image_no_pc_no_LRead ? ? ? ? H1 H10)
     | Exact (H3 ? ? H10)
     ]
  ].
Qed.

Lemma RAM_image_succ_LIn_ge_registers_no :
 (registers_no:nat ; r:register ; pc: nat ; m:message ; P,P':agent)
  (is_RAM_image_succ registers_no r pc P) ->
   (OStep P (LIn m) P') ->
    (ge m registers_no).
Proof.
 Intros;
 Generalize m P' H0;
 Clear m P' H0;
 Elim H;
  [ Intros;
    Inversion H0
  | Intros P0 Q i H0 i0 m P' H1;
    Inversion H1;
     [ Exact (H0 ? ? H7)
     | Apply (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? i0 H7)
     ]
  | Intros P0 Q i i0 H0 m P' H1;
    Inversion H1;
     [ Apply (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? i H7)
     | Exact (H0 ? ? H7)
     ]
  ].
Qed.

Lemma RAM_image_succ_LOut_lt_registers_no :
 (registers_no:nat ; r:register ; pc: nat ; m:message ; P,P':agent)
  (is_RAM_image_succ registers_no r pc P) ->
   (OStep P (LOut m) P') ->
    (lt m registers_no).
Proof.
 Intros;
 Generalize m P' H0;
 Clear m P' H0;
 Elim H;
  [ Intros;
    Inversion H0
  | Intros P0 Q i H0 i0 m P' H1;
    Inversion H1;
     [ Exact (H0 ? ? H7)
     | Exact (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? i0 H7)
     ]
  | Intros P0 Q i i0 H0 m P' H1;
    Inversion H1;
     [ Exact (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? i H7)
     | Exact (H0 ? ? H7)
     ]
  ].
Qed.

Lemma RAM_image_succ_Tau_RAM_image_out_pc:
 (registers_no:nat ; r:register ; pc:nat ; P:agent)
  (lt r registers_no) ->
   (is_RAM_image_succ registers_no r pc P) ->
    {P':agent |
     (is_RAM_image_out_pc registers_no (S pc) P') /\ (OStep P Tau P')}.
Proof.
 Intros registers_no r pc P H H0;
 Elim H0;
  [ Exists (AParallel (Message r) (Program (Action (Out (S pc)) Dead)));
    Split;
     [ Apply IRIOPCAParallel2;
        [ Constructor;
          Assumption
        | Constructor
        ]
     | Constructor
     ]
  | Intros P0 Q i H1 i0;
    Elim H1;
    Intros;
    Exists (AParallel x Q);
    Elim p;
    Intros;
    Split;
     [ Apply IRIOPCAParallel1;
        [ Assumption
        | Assumption
        ]
     | Apply SAParallelPL;
        [ Intro;
          Elim H4;
          Intros;
          Inversion H5
        | Assumption
        ]
     ]
  | Intros P0 Q i i0 H1;
    Elim H1;
    Intros;
    Exists (AParallel P0 x);
    Elim p;
    Intros;
    Split;
     [ Apply IRIOPCAParallel2;
        [ Assumption
        | Assumption
        ]
     | Apply SAParallelPR;
        [ Intro;
          Elim H4;
          Intros;
          Inversion H5
        | Assumption
        ]
     ]
  ].
Defined.

Lemma RAM_image_succ_Tau_deterministic:
 (registers_no:nat ; r:register ; pc:nat ; P,P',P'':agent)
  (is_RAM_image_succ registers_no r pc P) ->
   (OStep P Tau P') ->
    (OStep P Tau P'') ->
     P'=P''.
Proof.
 Intros;
 Generalize P' P'' H0 H1;
 Clear P' P'' H0 H1;
 Elim H;
  [ Clear H;
    Intros;
    Inversion H0;
    Inversion H1;
    Reflexivity
  | Clear H;
    Intros P0 Q i H i0 P' P'' H0 H1;
    Inversion H0;
     [ Inversion H1;
        [ Rewrite (H ? ? H7 H13);
          Reflexivity
        | Elim (stuck_RAM_image_no_pc ? ? i0);
          Exists P'1;
          Left;
          Assumption
        | Generalize (RAM_image_succ_LIn_ge_registers_no ? ? ? ? ? ? i H10);
          Intro Hge;
          Generalize (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? i0 H11);
          Intro Hlt;
          Elim (le_not_lt ? ? Hge Hlt)
        | Generalize (RAM_image_succ_LOut_lt_registers_no ? ? ? ? ? ? i H10);
          Intro Hlt;
          Generalize (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? i0 H11);
          Intro Hge;
          Elim (le_not_lt ? ? Hge Hlt)
        | Elim (RAM_image_succ_not_LRead ? ? ? ? ? ? i H10)
        | Elim (RAM_image_no_pc_no_LRead ? ? ? ? i0 H11)
        ]
     | Elim (stuck_RAM_image_no_pc ? ? i0);
       Exists P'0;
       Left;
       Assumption
     | Generalize (RAM_image_succ_LIn_ge_registers_no ? ? ? ? ? ? i H4);
       Intro Hge;
       Generalize (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? i0 H5);
       Intro Hlt;
       Elim (le_not_lt ? ? Hge Hlt)
     | Generalize (RAM_image_succ_LOut_lt_registers_no ? ? ? ? ? ? i H4);
       Intro Hlt;
       Generalize (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? i0 H5);
       Intro Hge;
       Elim (le_not_lt ? ? Hge Hlt)
     | Elim (RAM_image_succ_not_LRead ? ? ? ? ? ? i H4)
     | Elim (RAM_image_no_pc_no_LRead ? ? ? ? i0 H5)
     ]
  | Clear H;
    Intros P0 Q i i0 H P' P'' H0 H1;
    Inversion H0;
     [ Elim (stuck_RAM_image_no_pc ? ? i);
       Exists P'0;
       Left;
       Assumption
     | Inversion H1;
        [ Elim (stuck_RAM_image_no_pc ? ? i);
          Exists P'1;
          Left;
          Assumption
        | Rewrite (H ? ? H7 H13);
          Reflexivity
        | Generalize (RAM_image_succ_LOut_lt_registers_no ? ? ? ? ? ? i0 H11);
          Intro Hlt;
          Generalize (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? i H10);
          Intro Hge;
          Elim (le_not_lt ? ? Hge Hlt)
        | Generalize (RAM_image_succ_LIn_ge_registers_no ? ? ? ? ? ? i0 H11);
          Intro Hge;
          Generalize (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? i H10);
          Intro Hlt;
          Elim (le_not_lt ? ? Hge Hlt)
        | Elim (RAM_image_no_pc_no_LRead ? ? ? ? i H10)
        | Elim (RAM_image_succ_not_LRead ? ? ? ? ? ? i0 H11)
        ]
     | Generalize (RAM_image_succ_LOut_lt_registers_no ? ? ? ? ? ? i0 H5);
       Intro Hlt;
       Generalize (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? i H4);
       Intro Hge;
       Elim (le_not_lt ? ? Hge Hlt)
     | Generalize (RAM_image_succ_LIn_ge_registers_no ? ? ? ? ? ? i0 H5);
       Intro Hge;
       Generalize (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? i H4);
       Intro Hlt;
       Elim (le_not_lt ? ? Hge Hlt)
     | Elim (RAM_image_no_pc_no_LRead ? ? ? ? i H4)
     | Elim (RAM_image_succ_not_LRead ? ? ? ? ? ? i0 H5)
     ]
  ].
Qed.

(*** Lemmas about is_RAM_image_out_pc ***)

Lemma RAM_image_out_pc_not_LNot:
 (registers_no:nat ; pc:nat ; m:message ; P,P':agent)
  (is_RAM_image_out_pc registers_no pc P) ->
   ~(OStep P (LNot m) P').
Proof.
 Unfold not;
 Intros registers_no pc m P P' H H0;
 Generalize m P' H0;
 Elim H;
  [ Intros;
    Inversion H1
  | Intros P0 Q i H1 i0 m0 P'0 H2;
    Inversion H2;
     [ Apply H5;
       Exists m0;
       Reflexivity
     | Apply (RAM_image_no_pc_no_LNot ? ? ? ? i0 H8)
     | Exact (H1 ? ? H6)
     | Apply (RAM_image_no_pc_no_LNot ? ? ? ? i0 H6)
     ]
  | Intros P0 Q i i0 H1 m0 P'0 H2;
    Inversion H2;
     [ Apply (RAM_image_no_pc_no_LNot ? ? ? ? i H8)
     | Apply H5;
       Exists m0;
       Reflexivity
     | Apply (RAM_image_no_pc_no_LNot ? ? ? ? i H6)
     | Exact (H1 ? ? H6)
     ]
  ].
Qed.

Lemma RAM_image_out_pc_not_LRead:
 (registers_no:nat ; pc:nat ; m:message ; P,P':agent)
  (is_RAM_image_out_pc registers_no pc P) ->
   ~(OStep P (LRead m) P').
Proof.
 Unfold not;
 Intros;
 Generalize m P' H0;
 Elim H;
  [ Intros;
    Inversion H1
  | Intros P0 Q i H1 i0 m0 P'0 H2;
    Inversion H2;
     [ Exact (H1 ? ? H8)
     | Apply (RAM_image_no_pc_no_LRead ? ? ? ? i0 H8)
     ]
  | Intros P0 Q i i0 H1 m0 P'0 H2;
    Inversion H2;
     [ Apply (RAM_image_no_pc_no_LRead ? ? ? ? i H8)
     | Exact (H1 ? ? H8)
     ]
  ].
Qed.

Lemma RAM_image_out_pc_LIn_ge_registers_no :
 (registers_no:nat ; pc: nat ; m:message ; P,P':agent)
  (is_RAM_image_out_pc registers_no pc P) ->
   (OStep P (LIn m) P') ->
    (ge m registers_no).
Proof.
 Intros;
 Generalize m P' H0;
 Clear m P' H0;
 Elim H;
  [ Intros;
    Inversion H0
  | Intros P0 Q i H0 i0 m0 P'0 H1;
    Inversion H1;
     [ Exact (H0 ? ? H7)
     | Apply (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? i0 H7)
     ]
  | Intros P0 Q i i0 H0 m0 P'0 H1;
    Inversion H1;
     [ Apply (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? i H7)
     | Exact (H0 ? ? H7)
     ]
  ].
Qed.

Lemma RAM_image_out_pc_LOut_lt_registers_no :
 (registers_no:nat ; pc: nat ; m:message ; P,P':agent)
  (is_RAM_image_out_pc registers_no pc P) ->
   (OStep P (LOut m) P') ->
    (lt m registers_no).
Proof.
 Intros;
 Generalize m P' H0;
 Clear m P' H0;
 Elim H;
  [ Intros;
    Inversion H0
  | Intros P0 Q i H0 i0 m0 P'0 H1;
    Inversion H1;
     [ Exact (H0 ? ? H7)
     | Apply (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? i0 H7)
     ]
  | Intros P0 Q i i0 H0 m0 P'0 H1;
    Inversion H1;
     [ Apply (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? i H7)
     | Exact (H0 ? ? H7)
     ]
  ].
Qed.

Lemma RAM_image_out_pc_Tau_RAM_image_pc:
 (registers_no:nat ; pc:nat ; P:agent)
  (ge pc registers_no) ->
   (is_RAM_image_out_pc registers_no pc P) ->
    {P':agent |
     (OStep P Tau P') /\
      (exT ?
       [H:(is_RAM_image_pc registers_no P')]
        (pc_of_RAM_image_pc registers_no P' H)=pc)}.
Proof.
 Induction 2;
  [ Exists (AParallel (Message pc) (Program Dead));
    Split;
     [ Constructor
     | Cut
        (is_RAM_image_pc registers_no (AParallel (Message pc) (Program Dead)));
        [ Intro HEx;
          Exists HEx;
          Simpl;
          Elim (le_lt_dec registers_no pc);
           [ Reflexivity
           | Intro;
             Elim (le_not_lt ? ? H b)
           ]
        | Constructor;
           [ Constructor;
             Assumption
           | Constructor
           ]
        ]
     ]
  | Intros P0 Q i H1 i0;
    Elim H1;
    Intros;
    Elim p;
    Intros;
    Exists (AParallel x Q);
    Split;
     [ Apply SAParallelPL;
        [ Intro;
          Elim H4;
          Intros;
          Inversion H5
        | Assumption
        ]
     | Elim H3;
       Intros;
       Cut (is_RAM_image_pc registers_no (AParallel x Q));
        [ Intro HEx;
          Exists HEx;
          Rewrite <- H4;
          Apply pc_of_AParallel_equal_pc1
        | Apply IRIAParallel1;
           [ Assumption
           | Assumption
           ]
        ]
     ]
  | Intros P0 Q i i0 H1;
    Elim H1;
    Intros;
    Elim p;
    Intros;
    Exists (AParallel P0 x);
    Split;
     [ Apply SAParallelPR;
        [ Intro;
          Elim H4;
          Intros;
          Inversion H5
        | Assumption
        ]
     | Elim H3;
       Intros;
       Cut (is_RAM_image_pc registers_no (AParallel P0 x));
        [ Intro HEx;
          Exists HEx;
          Rewrite <- H4;
          Apply pc_of_AParallel_equal_pc2
        | Apply IRIAParallel2;
           [ Assumption
           | Assumption
           ]
        ]
     ]
  ].
Defined.

Lemma RAM_image_out_pc_Tau_deterministic:
 (registers_no:nat ; pc:nat ; P,P',P'':agent)
  (is_RAM_image_out_pc registers_no pc P) ->
   (OStep P Tau P') ->
    (OStep P Tau P'') ->
     P'=P''.
Proof.
 Intros;
 Generalize P' P'' H0 H1;
 Clear P' P'' H0 H1;
 Elim H;
  [ Clear H;
    Intros;
    Inversion H0;
    Inversion H1;
    Reflexivity
  | Clear H;
    Intros P0 Q i H i0 P' P'' H0 H1;
    Inversion H0;
     [ Inversion H1;
        [ Rewrite (H ? ? H7 H13);
          Reflexivity
        | Elim (stuck_RAM_image_no_pc ? ? i0);
          Exists P'1;
          Left;
          Assumption
        | Generalize (RAM_image_out_pc_LIn_ge_registers_no ? ? ? ? ? i H10);
          Intro Hge;
          Generalize (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? i0 H11);
          Intro Hlt;
          Elim (le_not_lt ? ? Hge Hlt)
        | Generalize (RAM_image_out_pc_LOut_lt_registers_no ? ? ? ? ? i H10);
          Intro Hlt;
          Generalize (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? i0 H11);
          Intro Hge;
          Elim (le_not_lt ? ? Hge Hlt)
        | Elim (RAM_image_out_pc_not_LRead ? ? ? ? ? i H10)
        | Elim (RAM_image_no_pc_no_LRead ? ? ? ? i0 H11)
        ]
     | Elim (stuck_RAM_image_no_pc ? ? i0);
       Exists P'0;
       Left;
       Assumption
     | Generalize (RAM_image_out_pc_LIn_ge_registers_no ? ? ? ? ? i H4);
       Intro Hge;
       Generalize (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? i0 H5);
       Intro Hlt;
       Elim (le_not_lt ? ? Hge Hlt)
     | Generalize (RAM_image_out_pc_LOut_lt_registers_no ? ? ? ? ? i H4);
       Intro Hlt;
       Generalize (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? i0 H5);
       Intro Hge;
       Elim (le_not_lt ? ? Hge Hlt)
     | Elim (RAM_image_out_pc_not_LRead ? ? ? ? ? i H4)
     | Elim (RAM_image_no_pc_no_LRead ? ? ? ? i0 H5)
     ]
  | Clear H;
    Intros P0 Q i i0 H P' P'' H0 H1;
    Inversion H0;
     [ Elim (stuck_RAM_image_no_pc ? ? i);
       Exists P'0;
       Left;
       Assumption
     | Inversion H1;
        [ Elim (stuck_RAM_image_no_pc ? ? i);
          Exists P'1;
          Left;
          Assumption
        | Rewrite (H ? ? H7 H13);
          Reflexivity
        | Generalize (RAM_image_out_pc_LOut_lt_registers_no ? ? ? ? ? i0 H11);
          Intro Hlt;
          Generalize (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? i H10);
          Intro Hge;
          Elim (le_not_lt ? ? Hge Hlt)
        | Generalize (RAM_image_out_pc_LIn_ge_registers_no ? ? ? ? ? i0 H11);
          Intro Hge;
          Generalize (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? i H10);
          Intro Hlt;
          Elim (le_not_lt ? ? Hge Hlt)
        | Elim (RAM_image_no_pc_no_LRead ? ? ? ? i H10)
        | Elim (RAM_image_out_pc_not_LRead ? ? ? ? ? i0 H11)
        ]
     | Generalize (RAM_image_out_pc_LOut_lt_registers_no ? ? ? ? ? i0 H5);
       Intro Hlt;
       Generalize (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? i H4);
       Intro Hge;
       Elim (le_not_lt ? ? Hge Hlt)
     | Generalize (RAM_image_out_pc_LIn_ge_registers_no ? ? ? ? ? i0 H5);
       Intro Hge;
       Generalize (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? i H4);
       Intro Hlt;
       Elim (le_not_lt ? ? Hge Hlt)
     | Elim (RAM_image_no_pc_no_LRead ? ? ? ? i H4)
     | Elim (RAM_image_out_pc_not_LRead ? ? ? ? ? i0 H5)
     ]
  ].
Qed.

(*** Lemmas about is_RAM_image_decjump ***)

Lemma RAM_image_decjump_message_multiplicity_O_LNot:
 (registers_no:nat ; r:message ; pc,pc':nat ; P:agent)
  (message_multiplicity r P)=O ->
   (is_RAM_image_decjump registers_no r pc' pc P) ->
    {P':agent |
     (is_RAM_image_out_pc registers_no pc' P') /\ (OStep P (LNot r) P')}.
Proof.
 Intros;
 Induction H0;
  [ Exists (Program (Action (Out pc') Dead));
    Split;
     [ Constructor
     | Constructor
     ]
  | Case HrecH0;
     [ Apply (message_multiplicty_AParallel_O_message_multiplicity_O_L r P0 Q H)
     | Intros x p;
       Case p;
       Intros;
       Exists (AParallel x Q);
       Split;
        [ Apply IRIOPCAParallel1;
           [ Assumption
           | Assumption
           ]
        | Apply SAParallelNL;
           [ Assumption
           | Generalize
              (message_multiplicty_AParallel_O_message_multiplicity_O_R
                r P0 Q H);
             Intro;
             Exact (message_multiplicity_O_NoMessage ? ? H4)
           ]
        ]
     ]
  | Case HrecH0;
     [ Apply (message_multiplicty_AParallel_O_message_multiplicity_O_R r P0 Q H)
     | Intros;
       Case a;
       Intros;
       Exists (AParallel P0 x);
       Split;
        [ Apply IRIOPCAParallel2;
           [ Assumption
           | Assumption
           ]
        | Apply SAParallelNR;
           [ Assumption
           | Generalize
              (message_multiplicty_AParallel_O_message_multiplicity_O_L
                r P0 Q H);
             Intro;
             Exact (message_multiplicity_O_NoMessage ? ? H4)
           ]
        ]
     ]
  ].
Defined.

Lemma RAM_image_decjump_LNot_deterministic:
 (registers_no:nat ; r:message ; pc,pc':nat ; P,P',P'':agent)
  (is_RAM_image_decjump registers_no r pc' pc P) ->
   (OStep P (LNot r) P') -> (OStep P (LNot r) P'') ->
    P'=P''.
Proof.
 Intros;
 Generalize P' P'' H0 H1;
 Clear P' P'' H0 H1;
 Elim H;
  [ Intros;
    Inversion H0;
    Inversion H1;
    Reflexivity
  | Intros P0 Q i H0 i0 P' P'' H1 H2;
    Inversion H1;
     [ Elim H5;
       Split with r;
       Reflexivity
     | Elim H5;
       Split with r;
       Reflexivity
     | Inversion H2;
        [ Elim H11;
          Split with r;
          Reflexivity
        | Elim H11;
          Split with r;
          Reflexivity
        | Rewrite (H0 ? ? H6 H12);
          Reflexivity
        | Elim (RAM_image_no_pc_no_LNot ? ? ? ? i0 H12)
        ]
     | Elim (RAM_image_no_pc_no_LNot ? ? ? ? i0 H6)
     ]
  | Intros P0 Q i i0 H0 P' P'' H1 H2;
    Inversion H1;
     [ Elim H5;
       Split with r;
       Reflexivity
     | Elim H5;
       Split with r;
       Reflexivity
     | Elim (RAM_image_no_pc_no_LNot ? ? ? ? i H6)
     | Inversion H2;
        [ Elim H11;
          Split with r;
          Reflexivity
        | Elim H11;
          Split with r;
          Reflexivity
        | Elim (RAM_image_no_pc_no_LNot ? ? ? ? i H12)
        | Rewrite (H0 ? ? H6 H12);
          Reflexivity
        ]
     ]
  ].
Qed.

Lemma RAM_image_decjump_LIn:
 (registers_no:nat ; r:message ; pc,pc':nat ; P:agent)
  (is_RAM_image_decjump registers_no r pc' pc P) ->
   {P':agent |
    (is_RAM_image_out_pc registers_no (S pc) P') /\ (OStep P (LIn r) P')}.
Proof.
 Induction 1;
  [ Exists (Program (Action (Out (S pc)) Dead));
    Split;
     [ Constructor
     | Constructor
     ]
  | Intros P0 Q i H0 i0;
    Case H0;
    Intros;
    Exists (AParallel x Q);
    Case a;
    Intros;
    Split;
     [ Apply IRIOPCAParallel1;
        [ Assumption
        | Assumption
        ]
     | Apply SAParallelPL;
        [ Intro;
          Case H3;
          Intros;
          Discriminate
        | Assumption
        ]
     ]
  | Intros P0 Q i i0 H0;
    Elim H0;
    Intros;
    Exists (AParallel P0 x);
    Case p;
    Intros;
    Split;
     [ Apply IRIOPCAParallel2;
        [ Assumption
        | Assumption
        ]
     | Apply SAParallelPR;
        [ Intro;
          Case H3;
          Intros;
          Discriminate
        | Assumption
        ]
     ]
  ].
Defined.

Lemma RAM_image_decjump_lt_r_registers_no_LIn_deterministic:
 (registers_no:nat ; r:message ; pc,pc':nat ; P,P',P'':agent)
  (lt r registers_no) ->
   (is_RAM_image_decjump registers_no r pc' pc P) ->
    (OStep P (LIn r) P') -> (OStep P (LIn r) P'') -> P'=P''.
Proof.
 Intros;
 Generalize P' P'' H1 H2;
 Clear P' P'' H1 H2;
 Elim H0;
  [ Intros;
    Inversion H1;
    Inversion H2;
    Reflexivity
  | Intros P0 Q i H1 i0 P' P'' H2 H3;
    Inversion H2;
     [ Inversion H3;
        [ Elim (H1 ? ? H9 H15);
          Reflexivity
        | Generalize (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? i0 H15);
          Intro;
          Elim (le_not_lt ? ? H16 H)
        ]
     | Generalize (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? i0 H9);
       Intro;
       Elim (le_not_lt ? ? H10 H)
     ]
  | Intros P0 Q i i0 H1 P' P'' H2 H3;
    Inversion H2;
     [ Generalize (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? i H9);
       Intro;
       Elim (le_not_lt ? ? H10 H)
     | Inversion H3;
        [ Generalize (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? i H15);
          Intro;
          Elim (le_not_lt ? ? H16 H)
        | Elim (H1 ? ? H9 H15);
          Reflexivity
        ]
     ]
  ].
Qed.

Lemma RAM_image_decjump_message_multiplicity_not_O_Tau:
 (registers_no:nat ; r:message ; pc,pc':nat ; P:agent)
  ~(message_multiplicity r P)=O ->
   (is_RAM_image_decjump registers_no r pc' pc P) ->
    {P':agent | (is_RAM_image_out_pc registers_no (S pc) P')/\(OStep P Tau P')}.
Proof.
 Intros;
 Induction H0;
  [ Elim H;
    Reflexivity
  | Generalize
     (not_message_multiplicty_AParallel_O_not_message_multiplicity r P0 Q H);
    Intro;
    Case H2;
     [ Intro;
       Elim (HrecH0 n);
       Intros;
       Case p;
       Intros;
       Exists (AParallel x Q);
       Split;
        [ Apply IRIOPCAParallel1;
           [ Assumption
           | Assumption
           ]
        | Apply SAParallelPL;
           [ Intro;
             Case H5;
             Intros;
             Discriminate
           | Assumption
           ]
        ]
     | Intro;
       Generalize
        (message_multiplicity_neq_O_RAM_image_no_pc_LOut registers_no r Q n H1);
       Intro;
       Case H3;
       Intros;
       Case a;
       Intros;
       Case H5;
       Intros;
       Generalize (RAM_image_decjump_LIn ? ? ? ? ? H0);
       Intro;
       Case H8;
       Intros;
       Case a0;
       Intros;
       Exists (AParallel x0 x);
       Split;
        [ Apply IRIOPCAParallel1;
           [ Assumption
           | Assumption
           ]
        | Apply (SAParallelTauIn1 P0 x0 Q x r);
           [ Assumption
           | Assumption
           ]
        ]
     ]
  | Generalize
     (not_message_multiplicty_AParallel_O_not_message_multiplicity r P0 Q H);
    Intro;
    Case H2;
     [ Intro;
       Generalize
        (message_multiplicity_neq_O_RAM_image_no_pc_LOut
          registers_no r P0 n H0);
       Intro;
       Case H3;
       Intros;
       Case a;
       Intros;
       Case H5;
       Intros;
       Generalize (RAM_image_decjump_LIn ? ? ? ? ? H1);
       Intro;
       Case H8;
       Intros;
       Case a0;
       Intros;
       Exists (AParallel x x0);
       Split;
        [ Apply IRIOPCAParallel2;
           [ Assumption
           | Assumption
           ]
        | Apply (SAParallelTauIn2 P0 x Q x0 r);
           [ Assumption
           | Assumption
           ]
        ]
     | Intro;
       Elim (HrecH0 n);
       Intros;
       Case p;
       Intros;
       Exists (AParallel P0 x);
       Split;
        [ Apply IRIOPCAParallel2;
           [ Assumption
           | Assumption
           ]
        | Apply SAParallelPR;
           [ Intro;
             Case H5;
             Intros;
             Discriminate
           | Assumption
           ]
        ]
     ]
  ].
Defined.

Lemma LNot_message_multiplicity_O:
 (r:message ; P,P':agent)
  (OStep P (LNot r) P') ->
   (message_multiplicity r P)=(0).
Proof.
 Induction P;
  [ Intros;
    Inversion H
  | Reflexivity
  | Intros;
    Simpl;
    Inversion H1;
     [ Elim H4;
       Exists r;
       Reflexivity
     | Elim H4;
       Exists r;
       Reflexivity
     | Rewrite (H ? H5);
       Rewrite (NoMessage_message_multiplicity_O ? ? H7);
       Reflexivity
     | Rewrite (H0 ? H5);
       Rewrite (NoMessage_message_multiplicity_O ? ? H7);
       Reflexivity
     ]
  ].
Qed.

Lemma RAM_image_decjump_Tau_not_message_multiplicity_O:
 (registers_no:nat ; r:message ; pc,pc':nat ; P,P':agent)
  (is_RAM_image_decjump registers_no r pc' pc P) ->
   (OStep P Tau P') ->
    ~(message_multiplicity r P)=(0).
Proof.
 Induction P;
  [ Intros;
    Inversion H
  | Intros;
    Intro;
    Inversion H;
    Rewrite <- H3 in H0;
    Inversion H0
  | Intros;
    Intro;
    Inversion H1;
     [ Inversion H2;
        [ Elim (H ? H6 H13);
          Apply
           (message_multiplicty_AParallel_O_message_multiplicity_O_L ? ? ? H3)
        | Apply (stuck_RAM_image_no_pc ? ? H7);
          Exists P'0;
          Left;
          Assumption
        | Generalize
           (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? H7 H11);
          Intro Hlt;
          Elim
           (RAM_image_decjump_LIn_eq_or_ge_registers_no ? ? ? ? ? ? ? H6 H10);
           [ Intro;
             Rewrite <- H13 in H3;
             Generalize
              (message_multiplicty_AParallel_O_message_multiplicity_O_R
                ? ? ? H3);
             Intro;
             Apply (LOut_not_message_multiplicity_eq_O ? ? ? H11 H14)
           | Intro Hge;
             Apply (le_not_lt ? ? Hge Hlt)
           ]
        | Generalize (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? H7 H11);
          Intro Hge;
          Generalize
           (RAM_image_decjump_LOut_lt_registers_no ? ? ? ? ? ? ? H6 H10);
          Intro Hlt;
          Apply (le_not_lt ? ? Hge Hlt)
        | Apply
           (RAM_image_decjump_not_LRead ? ? ? ? ? ? ? H6 H10)
        | Apply (RAM_image_no_pc_no_LRead ? ? ? ? H7 H11)
        ]
     | Inversion H2;
        [ Apply (stuck_RAM_image_no_pc ? ? H6);
          Exists P'0;
          Left;
          Assumption
        | Elim (H0 ? H7 H13);
          Apply
           (message_multiplicty_AParallel_O_message_multiplicity_O_R ? ? ? H3)
        | Generalize (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? H6 H10);
          Intro Hge;
          Generalize
           (RAM_image_decjump_LOut_lt_registers_no ? ? ? ? ? ? ? H7 H11);
          Intro Hlt;
          Apply (le_not_lt ? ? Hge Hlt)
        | Generalize
           (RAM_image_no_pc_LOut_lt_registers_no ? ? ? ? H6 H10);
          Intro Hlt;
          Elim
           (RAM_image_decjump_LIn_eq_or_ge_registers_no ? ? ? ? ? ? ? H7 H11);
           [ Intro;
             Rewrite <- H13 in H3;
             Generalize
              (message_multiplicty_AParallel_O_message_multiplicity_O_L
                ? ? ? H3);
             Intro;
             Apply (LOut_not_message_multiplicity_eq_O ? ? ? H10 H14)
           | Intro Hge;
             Apply (le_not_lt ? ? Hge Hlt)
           ]
        | Apply (RAM_image_no_pc_no_LRead ? ? ? ? H6 H10)
        | Apply
           (RAM_image_decjump_not_LRead ? ? ? ? ? ? ? H7 H11)
        ]
     ]
  ].
Qed.

Theorem RAM_image_decjump_Tau_not_LNot:
 (registers_no:nat ; r:message ; pc,pc':nat ; P,P',P'':agent)
  (is_RAM_image_decjump registers_no r pc' pc P) ->
   (OStep P Tau P') -> ~(OStep P (LNot r) P'').
Proof.
 Intros;
 Intro;
 Elim (eq_nat_dec (message_multiplicity r P) O);
  [ Intro;
    Apply (RAM_image_decjump_Tau_not_message_multiplicity_O ? ? ? ? ? ? H H0 a)
  | Intro;
    Apply b;
    Apply (LNot_message_multiplicity_O ? ? ? H1)
  ].
Qed.

(* We introduce the equivalence relation on RAM images that is *)
(* induced by the equality of the inverse images.              *)
Definition RAM_image_equivalence :
 (registers_no:nat ; P,P':agent)
  (is_RAM_image_pc registers_no P) ->
   (is_RAM_image_pc registers_no P') ->
    Prop
:=
 [registers_no:nat ; P,P':agent]
  [H:(is_RAM_image_pc registers_no P)]
   [H':(is_RAM_image_pc registers_no P')]
    (RAM_configuration_equality registers_no
     (configuration_of_RAM_image_pc registers_no P H)
     (configuration_of_RAM_image_pc registers_no P' H')).

Lemma is_RAM_image_out_pc_not_is_RAM_image_no_pc:
 (registers_no: nat ; r:message ; Q:agent)
  (is_RAM_image_out_pc registers_no r Q) ->
   ~(is_RAM_image_no_pc registers_no Q).
Proof.
 Induction 1;
  [ Intro;
    Inversion H0
  | Intros;
    Intro;
    Apply H1;
    Inversion H3;
    Assumption
  | Intros;
    Intro;
    Apply H2;
    Inversion H3;
    Assumption
  ].
Qed.

Lemma RAM_image_decjump_LIn_is_RAM_image_out_pc:
 (registers_no:nat ; r,m:message ; pc,pc':nat ; P,P':agent)
  (lt m registers_no) ->
   (is_RAM_image_decjump registers_no r pc' pc P) ->
    (OStep P (LIn m) P') -> (is_RAM_image_out_pc registers_no (S pc) P').
Proof.
 Intros;
 Generalize P' H1;
 Clear P' H1;
 Induction H0;
  [ Intros;
    Inversion H1;
    Constructor
  | Intros;
    Inversion H2;
     [ Apply IRIOPCAParallel1;
        [ Apply (HrecH0 ? H8)
	| Assumption
	]
     | Generalize (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? H1 H8);
       Intro Hge;
       Elim (le_not_lt ? ? Hge H)
     ]
  | Intros;
    Inversion H2;
     [ Generalize (RAM_image_no_pc_LIn_ge_registers_no ? ? ? ? H0 H8);
       Intro Hge;
       Elim (le_not_lt ? ? Hge H)
     | Apply IRIOPCAParallel2;
        [ Assumption
        | Apply (HrecH0 ? H8)
	]
     ]
  ].
Qed.
